from __future__ import annotations

from time import strftime
import numpy as np
from numpy.typing import NDArray
from ase.io.cube import write_cube

from ..density_matrices.frequency import FrequencyDensityMatrices
from ..density_matrices.base import WorkMetadata
from ..density_matrices.time import ConvolutionDensityMatrices
from ..utils import Result, get_gaussian_pulse_values
from ..voronoi import VoronoiWeights, EmptyVoronoiWeights
from .writer import Writer, ResultsCollector


class DensityWriter(Writer):

    """ Calculate density contributions

    Parameters
    ----------
    collector
        ResultsCollector object
    """

    def __init__(self,
                 collector: ResultsCollector):
        super().__init__(collector)
        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            self._ulm_tag = 'Time Density'
            assert len(self.density_matrices.pulses) == 1, 'Only one pulse allowed'
        else:
            assert isinstance(self.density_matrices, FrequencyDensityMatrices)
            self._ulm_tag = 'Frequency Density'

    @property
    def voronoi(self) -> VoronoiWeights:
        return EmptyVoronoiWeights()

    @property
    def common_arrays(self) -> dict[str, NDArray[np.float64] | NDArray[np.int64] | int | float]:
        from ..calculators import DensityCalculator

        common = super().common_arrays
        common.pop('eig_i')
        common.pop('eig_a')
        common.pop('eig_n')
        common.pop('imin')
        common.pop('imax')
        common.pop('amin')
        common.pop('amax')

        assert isinstance(self.calc, DensityCalculator)
        common['N_c'] = self.calc.N_c
        common['cell_cv'] = self.calc.cell_cv

        atoms = self.density_matrices.ksd.atoms
        common['atom_numbers_i'] = atoms.get_atomic_numbers()
        common['atom_positions_iv'] = atoms.get_positions()

        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            common['time_t'] = self.density_matrices.times * 1e-3
        else:
            assert isinstance(self.density_matrices, FrequencyDensityMatrices)
            common['freq_w'] = self.density_matrices.frequencies

        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            # If pulse is Gaussian pulse, then get dictionary with 'pulsefreq' and 'pulsefwhm'
            common.update(**get_gaussian_pulse_values(self.density_matrices.pulses[0]))

        return common

    def fill_ulm(self,
                 writer,
                 work: WorkMetadata,
                 result: Result):
        writer.fill(result['rho_g'])

    def write_empty_arrays_ulm(self, writer):
        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            Nt = len(self.density_matrices.times)
            writer.add_array('rho_tg', (Nt, ) + self.calc.gdshape, dtype=float)
        else:
            assert isinstance(self.density_matrices, FrequencyDensityMatrices)
            Nw = len(self.density_matrices.frequencies)
            writer.add_array('rho_wg', (Nw, ) + self.calc.gdshape, dtype=float)


def write_density(out_fname: str,
                  atoms,
                  data,
                  comment: str | None = None):
    """ Calculate density contribution and save in Gaussian cube file format.

    Parameters
    ----------
    out_fname
        File name of the resulting cube file.
    comment
        Comment line in the cube file.
    """

    if comment is None:
        comment = f'Cube file from rhodent, written on {strftime("%c")}'
    else:
        comment = comment.strip()

    assert data is not None
    with open(out_fname, 'w') as fp:
        write_cube(fp, atoms, data=data, comment=comment)



from __future__ import annotations

from typing import Generator
import numpy as np

from .base import RhoParameters
from rhodent.density_matrices.buffer import DensityMatrixBuffer
from rhodent.density_matrices.readers.gpaw import KohnShamRhoWfsReader
from rhodent.utils import safe_fill
from .base import BaseDistributor, RhoIndices
from ...typing import Communicator
from ...utils.memory import MemoryEstimate


class TimeDistributor(BaseDistributor):

    """ Iteratively read density matrices in the time domain

    This class uses the KohnShamRhoWfsReader to iteratively read wave functions
    (each rank reading the same times) and construct density matrices in the ground state
    Kohn-Sham basis for each time. The different ranks are reading different chunks of
    the density matrices. The density matrices are accumulated in a buffer and yielded
    when all times have been read.

    Parameters
    ----------
    rho_reader
        Density matrices reader
    comm
        Communicator
    log
        Logger
    """

    def __init__(self,
                 rho_reader: KohnShamRhoWfsReader,
                 parameters: RhoParameters | None = None,
                 comm: Communicator | None = None):
        if rho_reader.comm.size > 1:
            raise ValueError('Serial TimeDistributor must have serial reader.')

        super().__init__(rho_reader, parameters, comm=comm)

    @property
    def dtype(self):
        return np.float64

    @property
    def xshape(self):
        return (self.nt, )

    @property
    def dt(self):
        return self.rho_wfs_reader.dt

    @property
    def nt(self):
        return self.rho_wfs_reader.nt

    @property
    def time_t(self):
        return self.rho_wfs_reader.time_t

    def __str__(self) -> str:
        lines = []
        lines.append('Density matrices reader')
        lines.append('  Receiving density matrices in continuous chunks.')
        lines.append(f'    shape {self._parameters.nnshape}')
        lines.append(f'    received by {self.maxntimes} ranks')
        lines.append(f'  {self.niters} iterations to process all chunks')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        narrays = (2 if self.yield_re and self.yield_im else 1) * len(self.derivative_order_s)
        shape = self._parameters.nnshape + (self.nt, narrays)

        comment = f'Buffers hold {narrays} arrays ({self.describe_reim()})'
        memory_estimate = MemoryEstimate(comment=comment)
        memory_estimate.add_key('Density matrix buffers', shape, float,
                                on_num_ranks=self.comm.size)

        return memory_estimate

    def __iter__(self) -> Generator[DensityMatrixBuffer, None, None]:
        """ Yield density matrices for all times, chunk by chunk

        The wave function file is read in chunks, time by time, with
        the reading of times in the inner loop.

        Yields
        ------
        Chunks of the density matrix
        """
        read_dm = DensityMatrixBuffer(self._parameters.nnshape,
                                      (self.nt, ),
                                      np.float64)
        if self.yield_re:
            read_dm.zeros(True, 0)
        if self.yield_im:
            read_dm.zeros(False, 0)

        self.rho_wfs_reader.parallel_prepare()

        # Loop over the chunks this rank should gather
        for indices in self.work_loop(self.comm.rank):
            if indices is None:
                continue
            # Convert to reading indices
            n1 = slice(self._parameters.n1min + indices.n1.start, self._parameters.n1min + indices.n1.stop)
            n2 = slice(self._parameters.n2min + indices.n2.start, self._parameters.n2min + indices.n2.stop)
            gen = self.rho_wfs_reader.iread(indices.s, indices.k, n1, n2)
            for t in self.rho_wfs_reader.work_loop(self.rho_wfs_reader.comm.rank):
                if t is None:
                    continue
                dm_buffer = next(gen)
                for source_nn, dest_nn in zip(dm_buffer._iter_buffers(), read_dm[t]._iter_buffers()):
                    dest_nn[:] = source_nn

            yield read_dm

    @classmethod
    def from_reader(cls,
                    rho_nn_reader: KohnShamRhoWfsReader,
                    parameters: RhoParameters,
                    **kwargs) -> TimeDistributor:
        return cls(rho_nn_reader, parameters, **kwargs)


class AlltoallvTimeDistributor(TimeDistributor):

    """ Iteratively read density matrices in the time domain

    This class uses the KohnShamRhoWfsReader to iteratively read wave functions
    (one time per rank) and construct density matrices in the ground state Kohn-Sham
    basis for each time. When all ranks have read one time each, this class
    performs a redistribution of data, such that each rank only gets one chunk of the
    density matrices, but for all times. The density matrices are accumulated in a
    buffer and yielded when all times have been read.

    Parameters
    ----------
    rho_reader
        Density matrices reader
    """

    def __init__(self,
                 rho_reader: KohnShamRhoWfsReader,
                 parameters: RhoParameters | None = None):
        if rho_reader.lcao_rho_reader.striden != 0:
            raise ValueError('n stride must be 0 (index all) for alltoallv parallelized method')

        BaseDistributor.__init__(self, rho_reader, parameters, comm=rho_reader.comm)

    def __str__(self) -> str:
        nnshape = self.rho_wfs_reader.nnshape(*self.first_indices())

        lines = []
        lines.append('Parallel density matrices reader')
        lines.append('  Receiving density matrices in continuous chunks.')
        lines.append(f'    densiy matrix continous chunk {nnshape}')
        lines.append(f'    split into smaller chunks {self._parameters.nnshape}')
        lines.append(f'    received by {self.maxntimes} ranks')
        lines.append('  Redistributing into continuous time form.')
        lines.append(f'    sent to {self.maxnchunks} ranks')
        lines.append(f'  {self.niters} iterations to process all chunks')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        narrays = (2 if self.yield_re and self.yield_im else 1) * len(self.derivative_order_s)

        nnshape = self.rho_wfs_reader.nnshape(*self.first_indices())

        before_shape = self._parameters.nnshape + (self.maxnchunks, narrays)
        after_shape = self._parameters.nnshape + (self.nt, narrays)
        total_after_size = np.prod(self._parameters.nnshape + (self.nt, )) * self.maxnchunks * narrays

        shape = nnshape + (narrays, )

        comment = f'Buffers hold {narrays} arrays ({self.describe_reim()})'
        memory_estimate = MemoryEstimate(comment=comment)
        memory_estimate.add_key('Density matrix chunks', shape, float,
                                on_num_ranks=self.maxntimes)
        memory_estimate.add_key('Before parallel redistribution', before_shape, float,
                                on_num_ranks=self.maxntimes)
        memory_estimate.add_key('After parallel redistribution', after_shape, float,
                                total_size=total_after_size,
                                on_num_ranks=self.maxnchunks)

        return memory_estimate

    def first_indices(self):
        for first_indices_r in self.work_loop_by_ranks():
            break

        indices_by_rank = [chunk for chunk in first_indices_r if chunk is not None]
        indices_concat, _ = RhoIndices.concatenate_indices(indices_by_rank)
        return indices_concat

    def __iter__(self) -> Generator[DensityMatrixBuffer, None, None]:
        """ Yield density matrices for all times, chunk by chunk

        The wave function file is read in chunks, time by time. However,
        chunks are grouped together so that the density matrix at each
        time is read in large chunks. Each rank reads the same chunk for
        a different time. Then, the chunks and times are redistributed,
        so that each rank now holds a small chunk, but for many times.
        The same chunk is read for all times before it is yielded.

        Yields
        ------
        Chunks of the density matrix
        """
        log = self.log

        self.rho_wfs_reader.parallel_prepare()

        # Here x is a compound index for a slice of both n and M
        for chunki, chunks_r in enumerate(self.work_loop_by_ranks()):
            log.start('read_alltoallv')

            # The work this rank is supposed to read
            indices = chunks_r[self.comm.rank]
            indices_by_rank = [chunk for chunk in chunks_r if chunk is not None]

            # Number of chunks of nn-indices being read
            nchunks = len(indices_by_rank)
            assert nchunks > 0

            # Find out how much of the total density matrix need to be read to get only
            # the required chunks
            indices_concat, reduced_indices_by_rank = RhoIndices.concatenate_indices(indices_by_rank)

            if indices is None:
                # This rank does not want any slices of n1 and n2.
                # It will still potentially participate in the parallel reading of times
                assert self.comm.rank >= nchunks

            n1 = slice(self._parameters.n1min + indices_concat.n1.start,
                       self._parameters.n1min + indices_concat.n1.stop)
            n2 = slice(self._parameters.n2min + indices_concat.n2.start,
                       self._parameters.n2min + indices_concat.n2.stop)

            if self.comm.rank < self.maxntimes:
                # This rank will read
                contiguous_chunks_buffer = DensityMatrixBuffer(self._parameters.nnshape,
                                                               (nchunks, ),
                                                               np.float64)
            else:
                # This rank does not read any times
                contiguous_chunks_buffer = DensityMatrixBuffer(self._parameters.nnshape,
                                                               (0, ),
                                                               np.float64)
            if self.comm.rank < nchunks:
                # This rank will get a chunk of the density matrices after redistribution
                contiguous_time_buffer = DensityMatrixBuffer(self._parameters.nnshape,
                                                             (self.nt, ),
                                                             np.float64)
            else:
                contiguous_time_buffer = DensityMatrixBuffer(self._parameters.nnshape,
                                                             (0, ),
                                                             np.float64)
            if self.yield_re:
                contiguous_chunks_buffer.zeros(True, 0)
                contiguous_time_buffer.zeros(True, 0)
            if self.yield_im:
                contiguous_chunks_buffer.zeros(False, 0)
                contiguous_time_buffer.zeros(False, 0)

            gen = self.rho_wfs_reader.iread(indices_concat.s, indices_concat.k, n1, n2)

            for t_r in self.rho_wfs_reader.work_loop_by_ranks():
                # Number of times being read
                ntimes = sum(1 for t in t_r if t is not None)
                # Time index this rank is reading, or None if not reading
                globalt = t_r[self.comm.rank]

                # Read the density matrices for one time per rank,
                # with each rank reading a large chunk of the density matrix
                if globalt is not None:
                    read_dm = next(gen)

                    for recvrank, readindices in enumerate(reduced_indices_by_rank):
                        for source_nn, dest_nn in zip(read_dm._iter_buffers(),
                                                      contiguous_chunks_buffer[recvrank]._iter_buffers()):
                            safe_fill(dest_nn, source_nn[readindices.n1, readindices.n2])
                else:
                    # This rank actually has no time to read (number of times
                    # is not evenly divisible by number of ranks, and this rank
                    # is trying to read past the end)
                    # This rank will still participate in the alltoallv operation
                    assert self.comm.rank >= ntimes

                # Perform the redistributions, so that each rank now holds
                # a smaller chunk of the density matrix, but for many times.
                contiguous_chunks_buffer.redistribute(
                        contiguous_time_buffer, comm=self.comm,
                        source_indices_r=[(r, ) if r < nchunks else None for r in range(self.comm.size)],
                        target_indices_r=[None if t is None else (t, ) for t in t_r],
                        log=log if 0 in t_r else None)

            if self.comm.rank == 0:
                log(f'Chunk {chunki+1}/{self.niters}: Read and distributed density matrices in '
                    f'{log.elapsed("read_alltoallv"):.1f}s', who='Response', flush=True)

            if indices is not None:
                yield contiguous_time_buffer



from __future__ import annotations

from typing import Collection, Generator
import numpy as np
from numpy.typing import NDArray

from gpaw.tddft.units import au_to_as

from .base import BaseDistributor, RhoParameters
from .time import TimeDistributor, AlltoallvTimeDistributor
from ..buffer import DensityMatrixBuffer
from ..readers.gpaw import KohnShamRhoWfsReader
from ...utils import get_array_filter, safe_fill_larger, fast_pad
from ...perturbation import create_perturbation, PerturbationLike, PulsePerturbation
from ...utils.logging import format_times
from ...utils.memory import MemoryEstimate
from ...typing import Array1D


class PulseConvolver(BaseDistributor):
    r""" Class performing pulse convolution of density matrices.

    The procedure of the pulse convolution is the following:

    - The entire time series of (real and/or imaginary parts of) density matrices
      is read, in several chunks of indices (n1, n2). Each MPI rank works on
      different chunks.
    - Each chunk is Fourier transformed, divided by the Fourier transform of the
      original perturbation and multiplied by the Fourier transform of the
      new pulse(s).
    - Optionally, derivatives are computed by multiplying the density matrices
      in the frequency domain by factors of :math:`i \omega`.
    - Each chunk is inverse Fourier tranformed, and only selected times are kept.

    Additionally, this class can redistribute the resulting convoluted density matrices
    so that the each rank holds the entire density matrix, for a few times.

    Parameters
    ----------
    rho_nn_reader
        Object begin able to iteratively read density matrices in the time domain.
        Density matrices are split in chunks and distributed among ranks.
    perturbation
        The perturbation which the density matrices read by :attr:`rho_nn_reader`
        are a response to.
    pulses
        List of pulses to perform to convolution with.
    derivative_order_s
        List of derivative orders to compute.
    filter_times
        After convolution keep only these times (or the times closest to them).
        In atomic units.
    result_on_ranks
        List of ranks among which the resulting arrays will be distributed.
        Empty list (default) to distribute among all ranks.
    """

    def __init__(self,
                 rho_nn_reader: TimeDistributor,
                 perturbation: PerturbationLike,
                 pulses: Collection[PerturbationLike],
                 derivative_order_s: list[int] = [0],
                 filter_times: list[float] | Array1D[np.float64] | None = None,
                 result_on_ranks: list[int] = []):
        super().__init__(rho_nn_reader.rho_wfs_reader,
                         rho_nn_reader._parameters,
                         comm=rho_nn_reader.comm)
        self.rho_nn_reader = rho_nn_reader

        # Check if we need to perform upscaling
        wfs_time_t = self.rho_nn_reader.time_t
        dt = self.rho_nn_reader.dt
        self._warn_too_small_dt = False

        if filter_times is None or len(filter_times) < 2:
            self._upscaling = 1
        else:
            # See if there is a mismatch between wanted and existing time grids.
            # If so, then upscale the data in the Fourier transform step.
            self._requested_dt = min(np.diff(np.sort(filter_times)))
            upscaling = int(np.round(dt / self._requested_dt))
            if upscaling < 1:
                upscaling = 1
            elif upscaling > 100:
                self._warn_too_small_dt = True
                upscaling = 1
            self._upscaling = upscaling

        # Construct an upscaled times grid
        self._time_t = wfs_time_t[0] + dt/self.upscaling * np.arange(len(wfs_time_t) * self.upscaling)

        # And filter it
        self._flt_t = get_array_filter(self._time_t, filter_times)

        # Set up pulses and perturbation
        self.pulses = [create_perturbation(pulse) for pulse in pulses]
        if not all(isinstance(pulse, PulsePerturbation) for pulse in self.pulses):
            raise ValueError('Pulse convolution can only be performed with pulses of type PulsePerturbation.')
        self.perturbation = create_perturbation(perturbation)

        # Set up derivatives
        assert all(order in [0, 1, 2] for order in derivative_order_s)
        assert all(np.diff(derivative_order_s) > 0), 'Derivative orders must be strictly increasing'
        self.derivative_order_s = derivative_order_s

        # Check which ranks the result should be stored on
        if len(result_on_ranks) == 0:
            self._result_on_ranks = set(range(self.comm.size))
        else:
            assert all(isinstance(rank, int) and rank >= 0 and rank < self.comm.size
                       for rank in result_on_ranks), result_on_ranks
            self._result_on_ranks = set(result_on_ranks)

        self._dist_buffer: DensityMatrixBuffer | None = None

    @property
    def dtype(self):
        return np.float64

    @property
    def xshape(self):
        return (len(self.pulses), self.nt)

    @property
    def time_t(self) -> NDArray[np.float64]:
        """ Array of times corresponding to convoluted density matrices; in atomic units. """
        return self._time_t[self._flt_t]

    @property
    def nt(self) -> int:
        """ Number of times for which convoluted density matrices are calculated. """
        return len(self.time_t)

    @property
    def nlocalt(self) -> int:
        """ Number of times stored on this rank after redistribution of the result. """
        return (self.nt + self.nranks_result - 1) // self.nranks_result

    @property
    def upscaling(self) -> int:
        """ Upscaling factor.

        Data is upscaled in time by this factor during the Fourier transformation step,
        in order to calculate convoluted density matrices at a denser grid of times than
        what is present in the time-dependent wave functions file.
        """
        return self._upscaling

    @property
    def result_on_ranks(self) -> list[int]:
        """ Set of ranks among which the result will be distributed. """
        return sorted(self._result_on_ranks)

    @property
    def nranks_result(self) -> int:
        """ Number of ranks storing part of the result after redistribution. """
        return len(self._result_on_ranks)

    def distributed_work(self) -> list[list[int]]:
        # Empty list for ranks that will not have any part of the result
        timet_r = self.comm.size * [[]]
        for r, rank in enumerate(self.result_on_ranks):
            timet_r[rank] = list(range(r, self.nt, self.nranks_result))

        return timet_r

    def my_work(self) -> list[int]:
        timet_r = self.distributed_work()
        return timet_r[self.comm.rank]

    def __str__(self) -> str:
        wfs_nt = len(self.rho_nn_reader.time_t)
        dt = self.rho_nn_reader.dt

        lines = []
        lines.append('Pulse convolver')
        lines.append(f'  Performing convolution trick on {self.maxnchunks} ranks')
        lines.append('  Fast Fourier transform')
        lines.append(f'    matrix dimensions {self.rho_nn_reader._parameters.nnshape}')
        lines.append(f'    grid of {wfs_nt} times')
        lines.append(f'    {self.describe_reim()}')
        lines.append('   In frequency domain')
        lines.append(f'     calculating {self.describe_derivatives()}')
        if self._warn_too_small_dt:
            lines.append('WARNING:, the smallest spacing between requested times is ')
            lines.append(f'{self._requested_dt * au_to_as:.2f}. This is much smaller than the time step ')
            lines.append(f'in the time-dependent wave functions file ({dt * au_to_as:.2f} as). ')
            lines.append('No upscaling will be done.')
        elif self.upscaling == 1:
            lines.append('     not upscaling data')
        else:
            lines.append(f'     upscaling by factor {self.upscaling}')
            lines.append(f'       requested time step {self._requested_dt * au_to_as:.2f} as')
            lines.append(f'       time stpe in file {dt * au_to_as:.2f} as.')
        lines.append(f'     convolution with {len(self.pulses)} pulses')
        lines.append('  Fast inverse Fourier transform')
        lines.append(f'    keeping time grid of {self.nt} times')
        lines.append(f'      {format_times(self.time_t, units="au")}')
        lines.append('')

        lines.append('  Redistributing into full density matrices')
        lines.append(f'    {self.niters} iterations to process all chunks')
        lines.append(f'    matrix dimensions {self.rho_nn_reader._parameters.full_nnshape}')
        lines.append(f'    result stored on {self.nranks_result} ranks')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        parameters = self.rho_nn_reader._parameters

        narrays = (2 if self.yield_re and self.yield_im else 1) * len(self.derivative_order_s)
        temp_shape = parameters.nnshape + (self.maxnchunks, len(self.pulses), self.nlocalt, narrays)
        result_shape = parameters.full_nnshape + (len(self.pulses), self.nlocalt, narrays)

        total_result_size = int(np.prod(parameters.full_nnshape + (len(self.pulses), self.nt))) * narrays

        comment = f'Buffers hold {narrays} arrays ({self.describe_reim()}, {self.describe_derivatives()})'
        own_memory_estimate = MemoryEstimate(comment=comment)
        own_memory_estimate.add_key('Temporary buffer', temp_shape, float,
                                    on_num_ranks=self.nranks_result)
        own_memory_estimate.add_key('Result buffer', result_shape, float,
                                    total_size=total_result_size,
                                    on_num_ranks=self.nranks_result)

        memory_estimate = MemoryEstimate()
        memory_estimate.add_child('Time-dependent wave functions reader',
                                  self.rho_nn_reader.rho_wfs_reader.get_memory_estimate())
        memory_estimate.add_child('Parallel density matrices reader',
                                  self.rho_nn_reader.get_memory_estimate())
        memory_estimate.add_child('Pulse convolver',
                                  own_memory_estimate)

        return memory_estimate

    def _freq_domain_derivative(self,
                                order: int) -> NDArray[np.complex128 | np.float64]:
        r""" Take derivative in frequency space by multiplying by .. math:

        (i \omega)^n.

        Parameters
        ----------
        order
            Order :math:`n` of the derivative.
        """
        if order == 0:
            return np.array([1])

        padnt = fast_pad(self.rho_nn_reader.nt)
        dt = self.rho_nn_reader.dt
        omega_w = 2 * np.pi * np.fft.rfftfreq(padnt, dt)

        return (1.0j * omega_w) ** order  # type: ignore

    @property
    def dist_buffer(self) -> DensityMatrixBuffer:
        """ Buffer of denisty matrices on this rank after redistribution. """
        if self._dist_buffer is None:
            self._dist_buffer = self.redistribute()
        return self._dist_buffer

    def __iter__(self) -> Generator[DensityMatrixBuffer, None, None]:
        """ Iteratively read density matrices and perform the pulse convolution.

        Each iteration performs the calculation for a different chunk of the
        density matrix. Each rank works on a different set of chunks.
        All ranks always work on the entire grid of times, during all interations.

        Yields
        ------
        A chunk of the convoluted density matrices, for the requested times.
        """
        wfs_time_t = self.rho_nn_reader.time_t  # Times in wave functions file
        padnt = fast_pad(len(wfs_time_t))  # Pad with zeros

        # Take Fourier transform of pulses
        pulse_pt = [pulse.pulse.strength(wfs_time_t) for pulse in self.pulses]
        pulse_pw = np.fft.rfft(pulse_pt, axis=-1, n=padnt)

        # Create buffer for result
        dm_buffer = DensityMatrixBuffer(self.rho_nn_reader._parameters.nnshape,
                                        (len(self.pulses), self.nt),
                                        np.float64)
        dm_buffer.zero_buffers(real=self.yield_re, imag=self.yield_im, derivative_order_s=self.derivative_order_s)

        for read_buffer in self.rho_nn_reader:
            x = []
            if self.yield_re:
                x.append((read_buffer._re_buffers[0], dm_buffer._re_buffers))
            if self.yield_im:
                x.append((read_buffer._im_buffers[0], dm_buffer._im_buffers))
            for data_nnt, buffers in x:
                # Take the Fourier transform of the data (Rerho or Imrho) and divide by
                # the Fourier transform of the perturbation
                # The data is padded by zeros, circumventing the periodicity of the Fourier transform
                data_nnw = self.perturbation.normalize_frequency_response(data_nnt, wfs_time_t, padnt, axis=-1)

                # Loop over the desired outputs (and which derivative orders they are)
                for derivative, buffer_nnpt in buffers.items():
                    deriv_w = self._freq_domain_derivative(derivative)
                    for p, pulse_w in enumerate(pulse_pw):
                        # Multiply factor for derivative (power of I*omega)
                        # All timesteps cancel when taking fft->ifft, so do not scale by it
                        _data_nnw = data_nnw * (deriv_w * pulse_w)
                        # Inverse Fourier transform
                        # Optionally, the data is upscaled by padding with even more zeros
                        conv_nnt = np.fft.irfft(_data_nnw, n=padnt * self.upscaling, axis=-1) * self.upscaling
                        buffer_nnpt[..., p, :] = conv_nnt[..., :len(self._time_t)][..., self._flt_t]

            yield dm_buffer.copy()

    def create_out_buffer(self) -> DensityMatrixBuffer:
        """ Create the DensityMatrixBuffer to hold the temporary density matrix after each redistribution """
        parameters = self.rho_nn_reader._parameters
        nlocalt = self.nlocalt if self.comm.rank in self.result_on_ranks else 0
        out_dm = DensityMatrixBuffer(nnshape=parameters.nnshape,
                                     xshape=(self.maxnchunks, len(self.pulses), nlocalt),
                                     dtype=np.float64)
        out_dm.zero_buffers(real=self.yield_re, imag=self.yield_im, derivative_order_s=self.derivative_order_s)

        return out_dm

    def create_result_buffer(self) -> DensityMatrixBuffer:
        """ Create the DensityMatrixBuffer to hold the resulting density matrix """
        parameters = self.rho_nn_reader._parameters
        nnshape = parameters.full_nnshape
        full_dm = DensityMatrixBuffer(nnshape=nnshape,
                                      xshape=(len(self.pulses), len(self.my_work())),
                                      dtype=np.float64)
        full_dm.zero_buffers(real=self.yield_re, imag=self.yield_im, derivative_order_s=self.derivative_order_s)

        return full_dm

    def redistribute(self) -> DensityMatrixBuffer:
        """ Perform the pulse convolution and redistribute the resulting density matrices.

        During the pulse convolution step, the data is distributed such that each rank
        stores the entire time series for one chunk of the density matrices, i.e. indices n1, n2.

        This function then performs a redistribution of the data such that each rank stores full
        density matrices, for certain times.

        If the density matrices are split into more chunks than there are ranks, then the
        chunks are read, convoluted with pulses and distributed in a loop several times until all
        data has been processed.

        Returns
        -------
        Density matrix buffer with x-dimensions (number of pulses, number of local times)
        where the number of local times variers between the ranks.
        """
        local_work = iter(self)
        parameters = self.rho_nn_reader._parameters
        log = self.log
        self.rho_nn_reader.rho_wfs_reader.lcao_rho_reader.striden == 0, \
            'n stride must be 0 (index all) for redistribute'

        # Time indices of result on each rank
        timet_r = self.distributed_work()

        out_dm = self.create_out_buffer()
        full_dm = self.create_result_buffer()

        _exhausted = object()

        # Loop over the chunks of the density matrix
        for chunki, indices_r in enumerate(self.work_loop_by_ranks()):
            # At this point, each rank stores one unique chunk of the density matrix.
            # All ranks have the entire time series of data for their own chunk.
            # If there are more chunks than ranks, then this loop will run
            # for several iterations. If the number of chunks is not divisible by the number of
            # ranks then, during the last iteration, some of the chunks are None (meaning the rank
            # currently has no data).

            # List of chunks that each rank currently stores, where element r of the list
            # contains the chunk that rank r works with. Ranks higher than the length of the list
            # currently store no chunks.
            # The list itself is identical on all ranks.
            chunks_by_rank = [indices[2:] for indices in indices_r if indices is not None]

            ntargets = len(chunks_by_rank)

            if self.comm.rank < ntargets:
                # This rank has data to send. Compute the pulse convolution and store the result
                dm_buffer = next(local_work)
            else:
                # This rank has no data to send
                assert next(local_work, _exhausted) is _exhausted
                # Still, we need to create a dummy buffer
                dm_buffer = DensityMatrixBuffer(nnshape=parameters.nnshape,
                                                xshape=(0, 0), dtype=np.float64)
                dm_buffer.zero_buffers(real=self.yield_re, imag=self.yield_im,
                                       derivative_order_s=self.derivative_order_s)

            log.start('alltoallv')

            # Redistribute the data:
            # - dm_buffer stores single chunks of density matrices, for all times and pulses.
            # - out_dm will store several chunks, for a few times
            # source_indices_r describes which slices of dm_buffer should be sent to which rank
            # target_indices_r describes to which positions of the out_dm buffer should be received
            # from which rank
            source_indices_r = [None if len(t) == 0 else (slice(None), t) for t in timet_r]
            target_indices_r = [r if r < ntargets else None for r in range(self.comm.size)]
            dm_buffer.redistribute(out_dm,
                                   comm=self.comm,
                                   source_indices_r=source_indices_r,
                                   target_indices_r=target_indices_r,
                                   log=log)

            if self.comm.rank == 0:
                log(f'Chunk {chunki+1}/{self.niters}: distributed convoluted response in '
                    f'{log.elapsed("alltoallv"):.1f}s', who='Response', flush=True)

            # Copy the redistributed data into the aggregated results buffer
            for array_nnrpt, full_array_nnpt in zip(out_dm._iter_buffers(), full_dm._iter_buffers()):
                for r, nn_indices in enumerate(chunks_by_rank):
                    safe_fill_larger(full_array_nnpt[nn_indices], array_nnrpt[:, :, r])

        assert next(local_work, _exhausted) is _exhausted

        return full_dm

    @classmethod
    def from_reader(cls,  # type: ignore
                    rho_nn_reader: KohnShamRhoWfsReader,
                    parameters: RhoParameters,
                    perturbation: PerturbationLike,
                    pulses: Collection[PerturbationLike],
                    derivative_order_s: list[int] = [0],
                    filter_times: list[float] | Array1D[np.float64] | None = None,
                    result_on_ranks: list[int] = []) -> PulseConvolver:
        time_distributor = AlltoallvTimeDistributor(rho_nn_reader, parameters)
        pulse_convolver = cls(time_distributor,
                              pulses=pulses,
                              perturbation=perturbation,
                              derivative_order_s=derivative_order_s,
                              filter_times=filter_times,
                              result_on_ranks=result_on_ranks)
        return pulse_convolver



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from ..density_matrices.frequency import FrequencyDensityMatrices
from ..density_matrices.base import WorkMetadata
from ..density_matrices.time import ConvolutionDensityMatrices
from ..utils import Result, get_gaussian_pulse_values
from .writer import ResultsCollector, Writer


class DipoleWriter(Writer):

    """ Calculate dipole moment contributions, optionally broadened onto
    an energy grid as a transition contribution map

    Parameters
    ----------
    collector
        ResultsCollector object
    """

    def __init__(self,
                 collector: ResultsCollector,
                 only_one_pulse: bool):
        super().__init__(collector)
        self.only_one_pulse = only_one_pulse
        if only_one_pulse:
            if isinstance(self.density_matrices, ConvolutionDensityMatrices):
                assert len(self.density_matrices.pulses) == 1, 'Only one pulse allowed'
        else:
            if isinstance(self.density_matrices, ConvolutionDensityMatrices):
                self._ulm_tag = 'Time TCM'
            else:
                assert isinstance(self.density_matrices, FrequencyDensityMatrices)
                self._ulm_tag = 'TCM'

    @property
    def common_arrays(self) -> dict[str, NDArray[np.float64] | NDArray[np.int64] | int | float]:
        from ..calculators import DipoleCalculator
        assert isinstance(self.calc, DipoleCalculator)

        common = super().common_arrays

        if self.calc.sigma is not None:
            # There is an energy grid
            common['sigma'] = self.calc.sigma
            common['energy_o'] = np.array(self.calc.energies_occ)
            common['energy_u'] = np.array(self.calc.energies_unocc)

        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            common['time_t'] = self.density_matrices.times * 1e-3
        else:
            assert isinstance(self.density_matrices, FrequencyDensityMatrices)
            common['freq_w'] = self.density_matrices.frequencies
            common['frequency_broadening'] = self.density_matrices.frequency_broadening
            common['osc_prefactor_w'] = self.calc.oscillator_strength_prefactor

        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            # If pulses are Gaussian pulses, then get dictionaries of 'pulsefreq' and 'pulsefwhm'
            pulsedicts = [get_gaussian_pulse_values(pulse) for pulse in self.density_matrices.pulses]
            try:
                pulsefreqs = [d['pulsefreq'] for d in pulsedicts]
                pulsefwhms = [d['pulsefwhm'] for d in pulsedicts]

                if self.only_one_pulse:
                    common['pulsefreq'] = pulsefreqs[0]
                    common['pulsefwhm'] = pulsefwhms[0]
                else:
                    common['pulsefreq_p'] = np.array(pulsefreqs)
                    common['pulsefwhm_p'] = np.array(pulsefwhms)
            except KeyError:
                # Not GaussianPulses
                pass

        return common

    def fill_ulm(self,
                 writer,
                 work: WorkMetadata,
                 result: Result):
        if self.collector.calc_kwargs.get('yield_total_ou', False):
            writer.fill(result['dm_ouv'])

    def write_empty_arrays_ulm(self, writer):
        if not self.collector.calc_kwargs.get('yield_total_ou', False):
            return
        shape_ou = (len(self.calc.energies_occ), len(self.calc.energies_unocc))
        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            Nt = len(self.density_matrices.times)
            # Real dipole
            writer.add_array('dm_touv', (Nt, ) + shape_ou + (3, ), dtype=float)
        else:
            assert isinstance(self.density_matrices, FrequencyDensityMatrices)
            Nw = len(self.density_matrices.frequencies)
            # Complex polarizability
            writer.add_array('dm_wouv', (Nw, ) + shape_ou + (3, ), dtype=complex)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generator, Generic, NamedTuple, Iterable
from itertools import product, zip_longest

import numpy as np

from gpaw.mpi import world
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from ..buffer import DensityMatrixBuffer
from ..readers.gpaw import KohnShamRhoWfsReader
from ...utils import DTypeT, Logger, concatenate_indices, env
from ...typing import Communicator
from ...utils.memory import HasMemoryEstimate


class BaseDistributor(HasMemoryEstimate, ABC, Generic[DTypeT]):

    """ Distribute density matrices over time, frequency or other dimensions across MPI ranks
    """

    def __init__(self,
                 rho_reader: KohnShamRhoWfsReader,
                 parameters: RhoParameters | None = None,
                 comm: Communicator | None = None):
        self.rho_wfs_reader = rho_reader

        self._comm = world if comm is None else comm
        if parameters is None:
            parameters = RhoParameters.from_ksd(self.ksd, self.comm)
        self._parameters = parameters

        self.derivative_order_s = [0]

    @property
    @abstractmethod
    def dtype(self) -> np.dtype[DTypeT]:
        """ Dtype of buffers. """
        raise NotImplementedError

    @property
    @abstractmethod
    def xshape(self) -> tuple[int, ...]:
        """ Shape of x-dimension in buffers. """
        raise NotImplementedError

    @property
    def ksd(self) -> KohnShamDecomposition:
        """ Kohn-Sham decomposition object. """
        return self.rho_wfs_reader.ksd

    @property
    def comm(self) -> Communicator:
        """ MPI communicator. """
        return self._comm

    @property
    def yield_re(self) -> bool:
        """ Whether real part of density matrices is calculated. """
        return self.rho_wfs_reader.yield_re

    @property
    def yield_im(self) -> bool:
        """ Whether imaginary part of density matrices is calculated. """
        return self.rho_wfs_reader.yield_im

    @property
    def log(self) -> Logger:
        """ Logger object. """
        return self.rho_wfs_reader.log

    @abstractmethod
    def __iter__(self) -> Generator[DensityMatrixBuffer, None, None]:
        """ Yield density matrices in parts. Different data is
        yielded on different ranks

        Yields
        ------
        Part of the density matrix
        """
        raise NotImplementedError

    def work_loop(self,
                  rank: int) -> Generator[RhoIndices | None, None, None]:
        """ Like work_loop_by_rank but for one particular rank
        """
        for chunks_r in self.work_loop_by_ranks():
            yield chunks_r[rank]

    @property
    def niters(self) -> int:
        """ Number of iterations needed to read all chunks. """
        return len(list(self.work_loop_by_ranks()))

    @property
    def maxntimes(self) -> int:
        """ Maximum number of ranks participating in reading of times. """
        for t_r in self.rho_wfs_reader.work_loop_by_ranks():
            return sum(1 for t in t_r if t is not None)

        raise RuntimeError

    @property
    def maxnchunks(self) -> int:
        """ Maximum number of ranks participating in reading of chunks. """
        for chunks_r in self.work_loop_by_ranks():
            return sum(1 for chunk in chunks_r if chunk is not None)

        raise RuntimeError

    def describe_reim(self) -> str:
        if self.yield_re and self.yield_im:
            return 'Real and imaginary parts'
        elif self.yield_re:
            return 'Real part'
        else:
            return 'Imaginary part'

    def describe_derivatives(self) -> str:
        return 'derivative orders: ' + ', '.join([str(d) for d in self.derivative_order_s])

    def work_loop_by_ranks(self) -> Generator[list[RhoIndices | None], None, None]:
        """ Yield slice objects corresponding to the chunk of the density matrix
        that is gathered on each rank.

        New indices are yielded until the entire density matrix is processed
        (across all ranks).

        Yields
        ------
        List of slice objects corresponding to part of the density matrix
        yielded on each ranks.  None in place of the slice object if there is
        nothing yielded for that rank.
        """
        gen = self._parameters.iterate_indices()

        while True:
            chunks_r: list[RhoIndices | None] = [indices for _, indices
                                                 in zip(range(self.comm.size), gen)]

            remaining = self.comm.size - len(chunks_r)
            if remaining == 0:
                yield chunks_r
            elif remaining == self.comm.size:
                # There is nothing left to do for any rank
                break
            else:
                # Append Nones for the ranks that are not doing anything
                chunks_r += remaining * [None]
                yield chunks_r
                break

    def gather_on_root(self) -> Generator[DensityMatrixBuffer | None, None, None]:
        self.rho_wfs_reader.C0S_sknM   # Make sure to read this synchronously

        for indices_r, dm_buffer in zip_longest(self.work_loop_by_ranks(),
                                                self, fillvalue=None):
            assert indices_r is not None, 'Work loop shorter than work'

            # Yield root's own work
            if self.comm.rank == 0:
                assert indices_r[0] is not None
                assert dm_buffer is not None
                dm_buffer.ensure_contiguous_buffers()

                yield dm_buffer.copy()
            else:
                yield None

            # Yield the work of non-root
            for recvrank, recvindices in enumerate(indices_r[1:], start=1):
                if recvindices is None:
                    # No work on this recvrank
                    continue

                if self.comm.rank == 0:
                    # Receive work
                    assert dm_buffer is not None
                    dm_buffer.recv_arrays(self.comm, recvrank, log=self.log)
                    yield dm_buffer.copy()
                else:
                    # Send work to root if there is any
                    if self.comm.rank == recvrank:
                        assert dm_buffer is not None
                        dm_buffer.send_arrays(self.comm, 0, log=self.log)
                    yield None

    def collect_on_root(self) -> DensityMatrixBuffer | None:
        gen = self._parameters.iterate_indices()

        nnshape = (self._parameters.n1size, self._parameters.n2size)
        full_dm = DensityMatrixBuffer(nnshape, self.xshape, dtype=self.dtype)
        full_dm.zero_buffers(real=self.yield_re, imag=self.yield_im, derivative_order_s=self.derivative_order_s)

        for indices, dm_buffer in zip_longest(gen,
                                              self.gather_on_root(), fillvalue=None):
            if self.comm.rank != 0:
                continue

            assert indices is not None, 'Iterators must be same length'
            assert dm_buffer is not None, 'Iterators must be same length'

            s, k, n1, n2 = indices
            assert s == 0
            assert k == 0

            for partial_data, full_data in zip(dm_buffer._iter_buffers(), full_dm._iter_buffers()):
                _nn1, _nn2 = full_data[n1, n2].shape[:2]
                # Numpy struggles with the static type below
                full_data[n1, n2, :] += partial_data[:_nn1, :_nn2:]  # type: ignore
            self.log(f'Collected on root: density matrix slice [s={s}, k={k}, n1={n1}, n2={n2}].',
                     flush=True, who='Response')

        if self.comm.rank != 0:
            return None

        return full_dm

    @classmethod
    @abstractmethod
    def from_reader(cls,
                    rho_nn_reader: KohnShamRhoWfsReader,
                    parameters: RhoParameters,
                    **kwargs) -> BaseDistributor:
        """ Set up this class from a density matrix reader and parameters object

        """
        raise NotImplementedError

    @classmethod
    def from_parameters(cls,
                        wfs_fname: str,
                        ksd: KohnShamDecomposition | str,
                        comm=world,
                        yield_re: bool = True,
                        yield_im: bool = True,
                        stridet: int = 1,
                        log: Logger | None = None,
                        verbose: bool = False,
                        **kwargs):
        """ Set up this class, trying to enforce memory limit.

        Parameters
        ----------
        wfs_fname
            File name of the time-dependent wave functions file.
        ksd
            KohnShamDecomposition object or file name to the ksd file.
        comm
            MPI communicator.
        yield_re
            Whether to read and yield the real part of wave functions/density matrices.
        yield_im
            Whether to read and yield the imaginary part of wave functions/density matrices.
        stridet
            Skip this many steps when reading the time-dependent wave functions file.
        log
            Logger object.
        verbose
            Be verbose in the attempts to satisfy memory requirement.
        kwargs
            Options passed through to :func:`from_reader`.
        """
        # Set up the time-dependent wave functions reader which is always needed
        rho_reader = KohnShamRhoWfsReader(
            wfs_fname=wfs_fname, ksd=ksd, comm=comm,
            yield_re=yield_re, yield_im=yield_im, log=log, stridet=stridet)

        log = rho_reader.log

        # Get the target memory limit
        to_MiB = 1024 ** -2
        mem_limit = env.get_response_max_mem(comm.size) / to_MiB
        log('Attempting to set up response calculation with memory limit of '
            f'{mem_limit * to_MiB:.1f} MiB across all ranks.', who='Setup', rank=0)

        totals = []
        for iterations in range(1, 100):
            # Try setting up the distributor such that `iterations` iterations are
            # needed to process all chunks
            parameters = RhoParameters.from_ksd(rho_reader.ksd, comm, chunk_iterations=iterations)
            distributor = cls.from_reader(rho_reader, parameters, **kwargs)
            total = distributor.get_memory_estimate().grand_total
            totals.append(total)
            compare = totals[:-5:-1]  # Last 4 totals in reverse order
            last_changes = [tot_new / tot_old for tot_new, tot_old in zip(compare, compare[1:])]
            if len(last_changes) == 0:
                improvement = ''
            else:
                s = ', '.join([f'{(1 - change)*100:.1f}%' for change in last_changes])
                improvement = f'Last improvements {s}'

            if verbose:
                log(f'Trying splitting in {distributor.niters:3} chunks -- estimate {total * to_MiB:.1f} MiB. '
                    f'{improvement}', who='Setup', rank=0)
            if total < mem_limit:
                log(f'Found suitable set of parameters limiting the memory to {total * to_MiB:.1f} MiB.',
                    who='Setup', rank=0)
                return distributor
            if len(last_changes) == 3 and sum(last_changes) / 3 > 0.98:
                break

        parameters = RhoParameters.from_ksd(rho_reader.ksd, comm, chunk_iterations=iterations)
        distributor = cls.from_reader(rho_reader, parameters, **kwargs)
        total = distributor.get_memory_estimate().grand_total

        log(f'Cannot satisfy memory limit. Estimate is {total * to_MiB:.1f} MiB.',
            who='Setup', rank=0)

        return distributor


class RhoIndices(NamedTuple):

    s: int
    k: int
    n1: slice
    n2: slice

    @staticmethod
    def concatenate_indices(indices_list: Iterable[RhoIndices],
                            ) -> tuple[RhoIndices, list[RhoIndices]]:
        indices_list = list(indices_list)
        assert len(indices_list) > 0
        s, k = indices_list[0][:2]
        assert all(indices.s == s for indices in indices_list), f'All s must be identical {indices_list}'
        assert all(indices.k == k for indices in indices_list), f'All k must be identical {indices_list}'

        _indices_concat, _reduced_indices_list = concatenate_indices(
            [(indices.n1, indices.n2) for indices in indices_list])
        indices_concat = RhoIndices(s, k, *_indices_concat)
        reduced_indices_list = [RhoIndices(s, k, *indices) for indices in _reduced_indices_list]

        return indices_concat, reduced_indices_list


class RhoParameters(NamedTuple):

    """ Utility class to describe density matrix indices.

    Parameters
    ----------
    ns
        Number of spins.
    nk
        Number of kpoints.
    n1min
        Smallest index of row to read.
    n1max
        Largest index of row to read.
    n2min
        Smallest index of column to read.
    n2max
        Largest index of column to read.
    striden1
        Stride for reading rows. Each chunk will be this size in the first dimension.
    striden2
        Stride for reading columns. Each chunk will be this size in the second dimension.
    """

    ns: int
    nk: int
    n1min: int
    n1max: int
    n2min: int
    n2max: int
    striden1: int = 4
    striden2: int = 4

    def __post_init__(self):
        self.striden1 = min(self.striden1, self.n1size)
        self.striden2 = min(self.striden2, self.n2size)

    @property
    def full_nnshape(self) -> tuple[int, int]:
        """ Shape of the full density matrix to be read. """
        return (self.n1size, self.n2size)

    @property
    def nnshape(self) -> tuple[int, int]:
        """ Shape of each density matrix chunk. """
        return (self.striden1, self.striden2)

    @property
    def n1size(self) -> int:
        """ Size of full density matrix in the first dimension. """
        return self.n1max + 1 - self.n1min

    @property
    def n2size(self) -> int:
        """ Size of full density matrix in the first dimension. """
        return self.n2max + 1 - self.n2min

    def iterate_indices(self) -> Generator[RhoIndices, None, None]:
        """ Iteratively yield indices slicing chunks of the density matrix. """
        for s, k, n1, n2 in product(range(self.ns), range(self.nk),
                                    range(0, self.n1size, self.striden1),
                                    range(0, self.n2size, self.striden2)):
            indices = RhoIndices(s=0, k=0,
                                 n1=slice(n1, n1 + self.striden1),
                                 n2=slice(n2, n2 + self.striden2))
            yield indices

    @classmethod
    def from_ksd(cls,
                 ksd: KohnShamDecomposition,
                 comm: Communicator | None = None,
                 only_ia: bool = True,
                 chunk_iterations: int = 1,
                 **kwargs) -> RhoParameters:
        """ Initialize from KohnShamDecomposition.

        Parameters
        ----------
        ksd
            KohnShamDecomposition.
        comm
            MPI Communicator.
        only_ia
            ``True`` if the parameters should be set up such that
            the electron-hole part of the density matrix is read,
            otherwise full density matrix.
        chunk_iterations
            Attempt to set up the strides so that the total number of
            chunks is as close as possible but not more than the number
            of MPI ranks times :attr:`chunk_iterations`.
        kwargs
            Options passed through to the constructor.
        """
        if comm is None:
            comm = world

        # Number of spins, kpoints and states
        ns, nk, nn, _ = ksd.reader.proxy('C0_unM', 0).shape

        params = dict()
        if only_ia:
            # Dimensions of electron-hole part
            imin, imax, amin, amax = [int(i) for i in ksd.ialims()]

            params['n1min'], params['n2min'] = imin, amin
            params['n1max'], params['n2max'] = imax, amax
        else:
            params['n1min'], params['n2min'] = 0, 0
            params['n1max'], params['n2max'] = nn - 1, nn - 1

        # Set up a helper object get the size
        helper = cls(ns, nk, **params)

        # We want this many chunks in total
        target_nchunks = chunk_iterations * comm.size
        ar = helper.n2size / helper.n1size  # Aspect ratio of density matrix

        nsplits1 = max(int(np.floor(np.sqrt(target_nchunks / ar))), 1)
        nsplits2 = (target_nchunks + nsplits1 - 1) // nsplits1

        # Defaults
        params['striden1'] = (helper.n1size + nsplits1 - 1) // nsplits1
        params['striden2'] = (helper.n2size + nsplits2 - 1) // nsplits2

        # Overwrite the default options in params with explicitly set options
        params.update(**kwargs)

        return cls(ns, nk, **params)



from __future__ import annotations

from collections.abc import Sequence
import numpy as np
from typing import Any, Collection, Generator
from ase.units import Hartree, Bohr
from gpaw.tddft.units import as_to_au, au_to_eV, au_to_eA, eV_to_au

from .base import BaseObservableCalculator
from ..density_matrices.base import WorkMetadata
from ..density_matrices.time import ConvolutionDensityMatrixMetadata
from ..density_matrices.distributed.pulse import PulsePerturbation
from ..perturbation import PerturbationLike
from ..response import BaseResponse
from ..voronoi import VoronoiWeights
from ..typing import Array1D
from ..utils import ResultKeys, Result, broaden_n2e


class EnergyCalculator(BaseObservableCalculator):

    r""" Calculate energy contributions in the time domain.

    The total energy can be written

    .. math::

        E_\text{tot}(t) = E^{(0)}_\text{tot} + \sum_{ia}^\text{eh} E_{ia}(t) + E_\text{pulse}(t).

    The contributions to the total energy are

    .. math::

        E_{ia} = \frac{1}{2} \left[
        p_{ia}\dot{q}_{ia} - q_{ia} \dot{p}_{ia} - v_{ia} q_{ia} \right],

    the contributions to the Hartree-xc energy are

    .. math::

        E_{ia}^\text{Hxc} = -\frac{1}{2} \left[
        \omega_{ia} q_{ia}^2 - q_{ia} \dot{p}_{ia} - v_{ia} q_{ia} \right],

    and the rate of energy change is

    .. math::

        \dot{E}_{ia} = \frac{1}{2} \left[
        p_{ia}\ddot{q}_{ia} - q_{ia} \ddot{p}_{ia}
        - v_{ia} \dot{q}_{ia} - \dot{v}_{ia} q_{ia} \right].

    The matrix element :math:`v_{ia}` can be computed from the dipole matrix element

    .. math::

        \boldsymbol{\mu}_{ia} = \int \psi^{(0)}_i(\boldsymbol{r}) \boldsymbol{r}
        \psi^{(0)}_a(\boldsymbol{r}) \mathrm{d}\boldsymbol{r}

    projected on the direction of the perturbation :math:`\hat{\boldsymbol{e}}`,
    the occupation number difference :math:`f_{ia}` and
    the pulse amplitude :math:`v_\text{pulse}(t)`

    .. math::

        v_{ia} = \sqrt{2 f_{ia}}
        \boldsymbol{\mu}_{ia} \cdot\hat{\boldsymbol{e}}
        v_\text{pulse}.


    Parameters
    ----------
    response
        Response object.
    voronoi
        Voronoi weights object.
    filter_pair
        Filter electron-hole pairs (occupied-unoccupied transitions) in summation of energies.
        Provide a tuples (low, high) to compute the sum of energies
        :math:`E_{ia}` and :math:`E_{ia}^\text{Hxc}` for pairs with transition energy
        :math:`\varepsilon_a-\varepsilon_{i}` in the interval low-high (in units of eV).
    energies_occ
        Energy grid in units of eV for occupied levels (holes).
    energies_unocc
        Energy grid in units of eV for unoccupied levels (electrons).
    sigma
        Gaussian broadening width for energy grid in units of eV.
    times
        Compute energies in the time domain, for these times (or as close to them as possible).
        In units of as.
    pulses
        Compute energies in the time domain, in response to these pulses.
        If none, then no pulse convolution is performed.
    """
    def __init__(self,
                 response: BaseResponse,
                 voronoi: VoronoiWeights | None = None,
                 *,
                 filter_pair: tuple[float, float] = (0, np.inf),
                 energies_occ: list[float] | Array1D[np.float64] | None = None,
                 energies_unocc: list[float] | Array1D[np.float64] | None = None,
                 sigma: float | None = None,
                 times: list[float] | Array1D[np.float64] | None = None,
                 pulses: Collection[PerturbationLike] | None = None,
                 ):
        super().__init__(response=response,
                         voronoi=voronoi,
                         energies_occ=energies_occ,
                         energies_unocc=energies_unocc,
                         sigma=sigma,
                         times=times,
                         pulses=pulses)
        if len(filter_pair) != 2:
            raise ValueError('filter_pair must be tuple of two floats')
        self._filter_pair_low = float(filter_pair[0]) * eV_to_au
        self._filter_pair_high = float(filter_pair[1]) * eV_to_au

    def get_result_keys(self,
                        yield_total_E_ia: bool = False,
                        yield_proj_E_ia: bool = False,
                        yield_total_E_ou: bool = False,
                        yield_total_dists: bool = False,
                        direction: int | Sequence[int] = 2,
                        ) -> ResultKeys:
        r""" Get the keys that each result will contain, and dimensions thereof.

        Parameters
        ----------
        yield_total_E_ia
            The results should include the contributions in the electron-hole basis to
            the total energy :math:`E_{ia}` and Hartree-xc energy :math:`E_{ia}^\text{Hxc}`
        yield_proj_E_ia
            The results should include the contributions in the electron-hole basis to
            the total energy projected on the occupied and unoccupied Voronoi weights
            :math:`E_{ia} w_i` and :math:`E_{ia} w_a`.
        yield_total_E_ou
            The results should include the contributions the total energy broadened on the
            occupied and unoccupied energy grids
            :math:`\sum_{ia} E_{ia}\delta(\varepsilon_\text{occ}-\varepsilon_{i})
            \delta(\varepsilon_\text{unocc}-\varepsilon_{a})` and
        yield_total_dists
            The results should include the contributions the total energy and Hartree-xc energy
            broadened by electronic transition energy onto the unoccupied energies grid
            :math:`\sum_{ia} E_{ia} \delta(\varepsilon-\omega_{ia})` and
            :math:`\sum_{ia} E_{ia}^\text{Hxc} \delta(\varepsilon-\omega_{ia})`
        direction
            Direction :math:`\hat{\boldsymbol{e}}` of the polarization of the
            pulse. Integer 0, 1 or 2 to specify x, y or z, or the direction vector specified as
            a list of three values. Default: polarization along z.
        """
        nI = self.nproj
        imin, imax, amin, amax = self.ksd.ialims()
        ni = int(imax - imin + 1)
        na = int(amax - amin + 1)
        no = len(self.energies_occ)
        nu = len(self.energies_unocc)

        assert direction in [0, 1, 2] or (isinstance(direction, Sequence) and len(direction) == 3)

        resultkeys = ResultKeys('dm', 'total', 'total_Hxc', 'field',
                                'total_resonant', 'total_resonant_Hxc', 'Epulse')

        if yield_total_E_ia:
            resultkeys.add_key('E_ia', (ni, na))
            resultkeys.add_key('Ec_ia', (ni, na))
        if yield_total_dists:
            resultkeys.add_key('E_transition_u', nu)
            resultkeys.add_key('Ec_transition_u', nu)
        if yield_total_E_ou:
            resultkeys.add_key('E_ou', (no, nu))
            resultkeys.add_key('Ec_ou', (no, nu))

        resultkeys.add_key('total_proj_II', (nI, nI))
        resultkeys.add_key('total_Hxc_proj_II', (nI, nI))
        if yield_proj_E_ia:
            resultkeys.add_key('E_occ_proj_Iia', (nI, ni, na))
            resultkeys.add_key('E_unocc_proj_Iia', (nI, ni, na))

        return resultkeys

    def icalculate(self,
                   yield_total_E_ia: bool = False,
                   yield_proj_E_ia: bool = False,
                   yield_total_E_ou: bool = False,
                   yield_total_dists: bool = False,
                   direction: int | Sequence[int] = 2,
                   ) -> Generator[tuple[WorkMetadata, Result], None, None]:
        """ Iteratively calculate energies.

        Parameters
        ----------
        yield_total_E_ia
            The results should include the contributions in the electron-hole basis to
            the total energy :math:`E_{ia}` and Hartree-xc energy :math:`E_{ia}^\\text{Hxc}`
        yield_proj_E_ia
            The results should include the contributions in the electron-hole basis to
            the total energy projected on the occupied and unoccupied Voronoi weights
            :math:`E_{ia} w_i` and :math:`E_{ia} w_a`.
        yield_total_E_ou
            The results should include the contributions the total energy broadened on the
            occupied and unoccupied energy grids
            :math:`\\sum_{ia} E_{ia}\\delta(\\varepsilon_\\text{occ}-\\varepsilon_{i})
            \\delta(\\varepsilon_\\text{unocc}-\\varepsilon_{a})` and
        yield_total_dists
            The results should include the contributions the total energy and Hartree-xc energy
            broadened by electronic transition energy onto the unoccupied energies grid
            :math:`\\sum_{ia} E_{ia} \\delta(\\varepsilon-\\omega_{ia})` and
            :math:`\\sum_{ia} E_{ia}^\\text{Hxc} \\delta(\\varepsilon-\\omega_{ia})`
        direction
            Direction :math:`\\hat{\\boldsymbol{e}}` of the polarization of the
            pulse. Integer 0, 1 or 2 to specify x, y or z, or the direction vector specified as
            a list of three values. Default: polarization along z.

        Yields
        ------
        Tuple (work, result) on the root rank of the calculation communicator. \
        Does not yield on non-root ranks of the calculation communicator.

            work
                An object representing the metadata (time or pulse) for the work done.
            result
                Object containg the calculation results for this time and pulse.
        """
        include_energy_dists = (yield_total_dists or yield_total_E_ou)
        if include_energy_dists:
            assert self.sigma is not None

        assert direction in [0, 1, 2] or (isinstance(direction, Sequence) and len(direction) == 3)
        direction_v = np.zeros(3)
        if direction in [0, 1, 2]:
            direction_v[direction] = 1
        else:
            direction_v[:] = direction
            direction_v /= np.linalg.norm(direction_v)

        dm_p = direction_v @ self.ksd.dm_vp
        v0_p = dm_p * np.sqrt(2 * self.ksd.f_p)
        dm_ia = self.ksd.M_ia_from_M_p(dm_p)
        v0_ia = self.ksd.M_ia_from_M_p(v0_p)
        w_ia = self.ksd.M_ia_from_M_p(self.ksd.w_p)

        # List all keys that this method computes
        # This is necessary to safely send and receive data across ranks
        resultkeys = self.get_result_keys(yield_total_dists=yield_total_dists,
                                          yield_total_E_ia=yield_total_E_ia,
                                          yield_proj_E_ia=yield_proj_E_ia,
                                          yield_total_E_ou=yield_total_E_ou,
                                          )

        self._read_weights_diagonal()

        # Iterate over the pulses and times
        for work, dm in self.density_matrices:
            if dm.rank > 0:
                continue
            self.log.start('calculate')

            assert isinstance(work, ConvolutionDensityMatrixMetadata)
            assert isinstance(work.pulse, PulsePerturbation)
            pulsestr = work.pulse.pulse.strength(work.time * as_to_au)

            dipmom = - 2 * np.sum(dm_ia * dm.rho_ia.real)
            resonant_filter_ia = (w_ia > self._filter_pair_low) & (w_ia < self._filter_pair_high)

            Epulse = dipmom * pulsestr * au_to_eV

            # Calculate v_ia
            v_ia = v0_ia * pulsestr

            E_ia = -v_ia * dm.Q_ia
            E_ia -= dm.Q_ia * dm.dP_ia

            Ec_ia = E_ia.copy()

            E_ia += dm.P_ia * dm.dQ_ia
            E_ia *= 0.5 * au_to_eV

            Ec_ia -= w_ia * dm.Q_ia ** 2
            Ec_ia *= 0.5 * au_to_eV

            result = Result()
            if yield_total_E_ia:
                result['E_ia'] = E_ia
                result['Ec_ia'] = Ec_ia

            result['dm'] = dipmom * au_to_eA
            result['Epulse'] = Epulse
            result['field'] = pulsestr * Hartree / Bohr  # V/Å
            result['total'] = np.sum(E_ia)
            result['total_Hxc'] = np.sum(Ec_ia)
            result['total_resonant'] = E_ia.ravel() @ resonant_filter_ia.ravel()
            result['total_resonant_Hxc'] = Ec_ia.ravel() @ resonant_filter_ia.ravel()

            # (Optional) Broaden transitions by transition energy
            if yield_total_dists:
                assert self.sigma is not None
                result['E_transition_u'] = broaden_n2e(E_ia.ravel(), w_ia.ravel() * au_to_eV,
                                                       self.energies_unocc, self.sigma)
                result['Ec_transition_u'] = broaden_n2e(Ec_ia.ravel(), w_ia.ravel() * au_to_eV,
                                                        self.energies_unocc, self.sigma)

            # (Optional) Compute energy contribution matrix
            if yield_total_E_ou:
                result['E_ou'] = self.broaden_ia2ou(E_ia)
                result['Ec_ou'] = self.broaden_ia2ou(Ec_ia)

            # Initialize the remaining empty arrays
            result.create_all_empty(resultkeys)

            # Iterate over projections
            for iI, weight_n in enumerate(self._iterate_weights_diagonal):
                assert weight_n is not None
                weight_i = weight_n[self.flti]
                weight_a = weight_n[self.flta]

                for iI2, weight2_n in enumerate(self._iterate_weights_diagonal):
                    assert weight2_n is not None
                    weight2_a = weight2_n[self.flta]

                    result['total_proj_II'][iI, iI2] = np.einsum(
                        'ia,i,a->', E_ia, weight_i, weight2_a, optimize=True)
                    result['total_Hxc_proj_II'][iI, iI2] = np.einsum(
                        'ia,i,a->', Ec_ia, weight_i, weight2_a, optimize=True)

                if yield_proj_E_ia:
                    E_occ_proj_ia = E_ia * weight_i[:, None]
                    E_unocc_proj_ia = E_ia * weight_a[None, :]

                    result['E_occ_proj_Iia'][iI] = E_occ_proj_ia
                    result['E_unocc_proj_Iia'][iI] = E_unocc_proj_ia

            self.log_parallel(f'Calculated and broadened energies in {self.log.elapsed("calculate"):.2f}s '
                              f'for {work.desc}', flush=True)

            yield work, result

        if self.calc_comm.rank == 0:
            self.log_parallel('Finished calculating energies', flush=True)

    @property
    def _need_derivatives_real_imag(self) -> tuple[list[int], bool, bool]:
        return ([0, 1], True, True)

    @property
    def _voronoi_shapes(self) -> dict[str, tuple[int, ...]]:
        nI = self.voronoi.nproj
        if nI == 0:
            return {}
        Nn = self.voronoi.nn
        # Diagonal weights only
        return {'diagonal': (nI, Nn)}

    def calculate_and_write(self,
                            out_fname: str,
                            write_extra: dict[str, Any] = dict(),
                            include_tcm: bool = False,
                            save_dist: bool = False,
                            only_one_pulse: bool = True):
        """ Calculate energy contributions.

        Energies are saved in a numpy archive if the file extension is `.npz`
        or in an ULM file if the file extension is `.ulm`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of extra key-value pairs to write to the data file.
        include_tcm
            Whether the transition contribution map (TCM) to the energies should be computed and saved.
        save_dist
            Whether the transition energy distributions should be computed and saved.
        only_one_pulse
            If False, group arrays by pulse.
        """
        from ..writers.energy import EnergyWriter
        from ..writers.writer import TimeResultsCollector, PulseConvolutionResultsCollector

        out_fname = str(out_fname)
        if out_fname[-4:] == '.npz':
            calc_kwargs = dict(yield_total_E_ou=include_tcm, yield_total_dists=save_dist)
            cls = TimeResultsCollector if only_one_pulse else PulseConvolutionResultsCollector
            writer = EnergyWriter(cls(self, calc_kwargs), only_one_pulse=only_one_pulse)
            writer.calculate_and_save_npz(out_fname, write_extra=write_extra)
        elif out_fname[-4:] == '.ulm':
            assert only_one_pulse
            calc_kwargs = dict(yield_total_E_ou=include_tcm, yield_total_dists=save_dist)
            writer = EnergyWriter(TimeResultsCollector(self, calc_kwargs, exclude=['E_ou']), only_one_pulse=True)
            writer.calculate_and_save_ulm(out_fname, write_extra=write_extra)
        else:
            raise ValueError(f'output-file must have ending .npz or .ulm, is {out_fname}')



from __future__ import annotations
from typing import Union, TypeVar

import numpy as np
from numpy.typing import NDArray

from gpaw import mpi
from gpaw.calculator import GPAW

GPAWCalculator = GPAW
Communicator = mpi._Communicator
ArrayIndex = Union[NDArray[np.int_], list[int], int, slice]


DTypeT = TypeVar('DTypeT', bound=np.generic, covariant=True)


class ArrayIsOnRootRank(NDArray[DTypeT]):
    def __new__(cls):
        """ Instances will act as empty numpy arrays """
        return NDArray.__new__(cls, (0, ))


DistributedArray = Union[NDArray[DTypeT], ArrayIsOnRootRank]
Array1D = np.ndarray[tuple[int], np.dtype[DTypeT]]
Array2D = np.ndarray[tuple[int, int], np.dtype[DTypeT]]
Array3D = np.ndarray[tuple[int, int, int], np.dtype[DTypeT]]



from __future__ import annotations

from typing import cast
import numpy as np
from numpy.typing import NDArray
from gpaw.mpi import world

from ..density_matrices.time import ConvolutionDensityMatrices
from .writer import (Writer, ResultsCollector, TimeAverageResultsCollector,
                     PulseConvolutionAverageResultsCollector)
from ..utils import get_gaussian_pulse_values


class HotCarriersWriter(Writer):

    """ Calculate hot-carrier totals, and optionally broadened hot-carrier energy distributions

    Parameters
    ----------
    collector
        ResultsCollector object
    only_one_pulse
        False if the resulting outputs should have one dimension corresponding
        to different pulses. True if there should be no such dimension. If True,
        then the calculator must only hold one pulse.
    """

    def __init__(self,
                 collector: ResultsCollector,
                 only_one_pulse: bool):
        super().__init__(collector)
        self.only_one_pulse = only_one_pulse
        if only_one_pulse:
            if isinstance(self.density_matrices, ConvolutionDensityMatrices):
                assert len(self.density_matrices.pulses) == 1, 'Only one pulse allowed'
        else:
            assert isinstance(self.density_matrices, ConvolutionDensityMatrices)

    @property
    def common_arrays(self) -> dict[str, NDArray[np.float64] | NDArray[np.int64] | int | float]:
        common = super().common_arrays
        assert isinstance(self.density_matrices, ConvolutionDensityMatrices)

        if isinstance(self.collector, (TimeAverageResultsCollector, PulseConvolutionAverageResultsCollector)):
            # Averages over time are taken
            common['avgtime_t'] = self.density_matrices.times * 1e-3
        else:
            common['time_t'] = self.density_matrices.times * 1e-3

        if isinstance(self.density_matrices, ConvolutionDensityMatrices):
            # If pulses are Gaussian pulses, then get dictionaries of 'pulsefreq' and 'pulsefwhm'
            pulsedicts = [get_gaussian_pulse_values(pulse) for pulse in self.density_matrices.pulses]

            try:
                pulsefreqs = [d['pulsefreq'] for d in pulsedicts]
                pulsefwhms = [d['pulsefwhm'] for d in pulsedicts]

                if self.only_one_pulse:
                    common['pulsefreq'] = pulsefreqs[0]
                    common['pulsefwhm'] = pulsefwhms[0]
                else:
                    common['pulsefreq_p'] = np.array(pulsefreqs)
                    common['pulsefwhm_p'] = np.array(pulsefwhms)
            except KeyError:
                # Not GaussianPulses
                pass

        if self.calc.sigma is not None:
            # There is an energy grid
            common['sigma'] = self.calc.sigma
            common['energy_o'] = np.array(self.calc.energies_occ)
            common['energy_u'] = np.array(self.calc.energies_unocc)

        return common


def write_hot_carrier_distributions(out_fname: str,
                                    writer: HotCarriersWriter):
    zerostr = 'relative to Fermi level'
    average_times = isinstance(writer.collector, (TimeAverageResultsCollector, PulseConvolutionAverageResultsCollector))

    energies_occ = writer.calc.energies_occ
    energies_unocc = writer.calc.energies_unocc
    if len(energies_occ) == 0 and len(energies_unocc) == 0:
        raise ValueError('Either occupied or unoccupied energies grid must be given')

    _data = dict(**writer.common_arrays)
    _data.update(writer.calculate_data()._data)

    if world.rank > 0:
        return

    pulsefreq = _data.pop('pulsefreq', None)
    pulsefwhm = _data.pop('pulsefwhm', None)

    data = cast(dict[str, NDArray[np.float64]], _data)

    # Set up array to be written in the data file.
    # Rows are energies, columns are projections and/or times
    nI = len(writer.voronoi)
    ne = max(len(energies_occ), len(energies_unocc))  # Longest length of energies

    eh_labels = []
    if len(energies_occ) > 0:
        # Compute hole distributions
        eh_labels.append('H')
    if len(energies_unocc) > 0:
        # Compute electron distributions
        eh_labels.append('E')

    if average_times:
        nt = 1
    else:
        nt = len(data['time_t'])

    ncolspertime = len(eh_labels) * (1 + nI)

    savedata = np.full((ne, len(eh_labels) + nt * ncolspertime), np.nan)
    savedata_by_times = [savedata[:, col:col + ncolspertime]
                         for col in range(len(eh_labels), savedata.shape[1], ncolspertime)]

    # Set up format string
    fmt = len(eh_labels) * ['%15.6f'] + nt * ncolspertime * ['%18.8e']

    # Set up header contents
    header_lines = [f'Hot carrier (H=hole, E=electron) distributions {zerostr}']

    if pulsefreq is not None:
        header_lines.append(f'Response to pulse with frequency {pulsefreq:.2f}eV, '
                            f'FWHM {pulsefwhm:.2f}fs')

    if average_times:
        avgtimes = data['avgtime_t']
        header_lines.append(f'Averaged for {len(avgtimes)} times between '
                            f'{avgtimes[0]:.1f}fs-{avgtimes[-1]:.1f}fs')
    else:
        times = data['time_t']
        header_lines.append(f'Computed for the following {len(times)} times (in units of fs)')
        header_lines += [f'  {time:.4f}' for t, time in enumerate(times)]

    if nI > 0:
        header_lines.append('Atomic projections')
        header_lines += [f'  {i:4.0f}: {str(proj)}' for i, proj in enumerate(writer.voronoi.atom_projections)]

    header_lines.append(f'Gaussian folding, Width {writer.calc.sigma:.4f}eV')
    desc_entries = ([f'{label} energy (eV)' for label in eh_labels] +
                    [f'Total {label} (1/eV)' for label in eh_labels] +
                    [f'Proj {s} {i:2.0f} (1/eV)' for i in range(nI) for s in eh_labels])
    desc_entries = ([f'{s:>15}' for s in desc_entries[:len(eh_labels)]] +
                    [f'{s:>18}' for s in desc_entries[len(eh_labels):]])
    desc_entries[0] = desc_entries[0][2:]  # Remove two spaces to account for '# '
    if not average_times:
        desc_entries.append(' ... repeated for next times')
    header_lines.append(' '.join(desc_entries))

    # Write the data to the array
    if average_times:
        if len(eh_labels) == 2:
            # Computed both electron and hole distributions
            savedata[:len(energies_occ), 0] = energies_occ
            savedata[:len(energies_unocc), 1] = energies_unocc

            savedata[:len(energies_occ), 2] = data['hcdist_o']
            savedata[:len(energies_unocc), 3] = data['hcdist_u']

            if nI > 0:
                savedata[:len(energies_occ), 4::2] = data['hcdist_proj_Io'].T
                savedata[:len(energies_unocc), 5::2] = data['hcdist_proj_Iu'].T
        elif 'H' in eh_labels:
            # Only hole distributions
            savedata[:, 0] = energies_occ
            savedata[:, 1] = data['hcdist_o']

            if nI > 0:
                savedata[:, 2:] = data['hcdist_proj_Io'].T
        else:
            # Only electron distributions
            savedata[:, 0] = energies_unocc
            savedata[:, 1] = data['hcdist_u']

            if nI > 0:
                savedata[:, 1:] = data['hcdist_proj_Iu'].T
    else:
        if len(eh_labels) == 2:
            # Computed both electron and hole distributions
            savedata[:len(energies_occ), 0] = energies_occ
            savedata[:len(energies_unocc), 1] = energies_unocc

            for t, sdata in enumerate(savedata_by_times):
                sdata[:len(energies_occ), 0] = data['hcdist_to'][t]
                sdata[:len(energies_unocc), 1] = data['hcdist_tu'][t]

                if nI > 0:
                    sdata[:len(energies_occ), 2::2] = data['hcdist_proj_tIo'][t].T
                    sdata[:len(energies_unocc), 3::2] = data['hcdist_proj_tIu'][t].T
        elif 'H' in eh_labels:
            # Only hole distributions
            savedata[:, 0] = energies_occ

            for t, sdata in enumerate(savedata_by_times):
                sdata[:, 0] = data['hcdist_to'][t]

                if nI > 0:
                    sdata[:, 1:] = data['hcdist_proj_tIo'][t].T
        else:
            # Only electron distributions
            savedata[:, 0] = energies_unocc

            for t, sdata in enumerate(savedata_by_times):
                sdata[:, 0] = data['hcdist_tu'][t]

                if nI > 0:
                    sdata[:, 1:] = data['hcdist_proj_tIu'][t].T

    np.savetxt(out_fname, savedata, fmt, header='\n'.join(header_lines))
    writer.calc.log(f'Written {out_fname}', who='Calculator', flush=True)


def write_hot_carrier_totals_by_pulse(out_fname: str,
                                      writer: HotCarriersWriter):

    Np = len(writer.calc.pulses)
    data = dict(**writer.common_arrays)
    data.update(writer.calculate_data()._data)

    if world.rank > 0:
        return

    nI = len(writer.voronoi)
    avgtimes = data['avgtime_t']
    assert isinstance(avgtimes, np.ndarray)

    savedata = np.full((Np, 2*(2+nI)), np.nan)
    savedata[:, 0] = data['pulsefreq_p']
    savedata[:, 1] = data['pulsefwhm_p']
    savedata[:, 2] = data['sumocc_p']
    savedata[:, 3] = data['sumunocc_p']
    savedata[:, 4::2] = data['sumocc_proj_pI']
    savedata[:, 5::2] = data['sumunocc_proj_pI']

    projectionsstr = '\n'.join([f'  {i:4.0f}: {str(proj)}'
                                for i, proj in enumerate(writer.voronoi.atom_projections)])
    desc_entries = (['Pulse freq (eV)', 'Pulse FWHM (fs)', 'Total H', 'Total E'] +
                    [f'Proj {s} {i:2.0f}' for i in range(nI) for s in 'HE'])
    desc_entries = ([f'{s:>17}' for s in desc_entries[:2]] +
                    [f'{s:>15}' for s in desc_entries[2:]])
    desc_entries[0] = desc_entries[0][2:]  # Remove two spaces to account for '# '

    header = (f'Hot carrier (H=hole, E=electron) numbers\n'
              f'Averaged for {len(avgtimes)} times between '
              f'{avgtimes[0]:.1f}fs-{avgtimes[-1]:.1f}fs\n'
              'Atomic projections:\n'
              f'{projectionsstr}\n'
              f'{" ".join(desc_entries)}')
    fmt = 2*['%17.6f'] + (2*(nI + 1))*['%15.8e']
    np.savetxt(out_fname, savedata, fmt, header=header)
    writer.calc.log(f'Written {out_fname}', who='Calculator', flush=True)


def write_hot_carrier_totals_by_time(out_fname: str,
                                     writer: HotCarriersWriter):
    Nt = len(writer.calc.times)
    data = dict(**writer.common_arrays)
    data.update(writer.calculate_data()._data)

    if world.rank > 0:
        return

    nI = len(writer.voronoi)

    savedata = np.full((Nt, 1 + 2*(1+nI)), np.nan)
    savedata[:, 0] = data['time_t']
    savedata[:, 1] = data['sumocc_t']
    savedata[:, 2] = data['sumunocc_t']
    savedata[:, 3::2] = data['sumocc_proj_tI']
    savedata[:, 4::2] = data['sumunocc_proj_tI']

    projectionsstr = '\n'.join([f'  {i:4.0f}: {str(proj)}'
                                for i, proj in enumerate(writer.voronoi.atom_projections)])
    desc_entries = (['Time (fs)', 'Total H', 'Total E'] +
                    [f'Proj {s} {i:2.0f}' for i in range(nI) for s in 'HE'])
    desc_entries = ([f'{s:>17}' for s in desc_entries[:1]] +
                    [f'{s:>15}' for s in desc_entries[1:]])
    desc_entries[0] = desc_entries[0][2:]  # Remove two spaces to account for '# '

    header_lines = ['Hot carrier (H=hole, E=electron) numbers\n']
    if 'pulsefreq' in data:
        header_lines += [f'Response to pulse with frequency {data["pulsefreq"]:.2f}eV, '
                         f'FWHM {data["pulsefwhm"]:.2f}fs']
    header_lines += ['Atomic projections:',
                     f'{projectionsstr}\n',
                     ' '.join(desc_entries)]
    header = '\n'.join(header_lines)
    fmt = ['%17.6f'] + (2*(nI + 1))*['%15.8e']
    np.savetxt(out_fname, savedata, fmt, header=header)
    writer.calc.log(f'Written {out_fname}', who='Calculator', flush=True)



from __future__ import annotations

from typing import Generator
import numpy as np

from gpaw.tddft.units import au_to_eV

from .base import BaseDistributor, RhoParameters
from .time import TimeDistributor, AlltoallvTimeDistributor
from ..buffer import DensityMatrixBuffer
from ..readers.gpaw import KohnShamRhoWfsReader
from ...utils import get_array_filter, safe_fill_larger, fast_pad
from ...utils.logging import format_frequencies
from ...utils.memory import MemoryEstimate
from ...perturbation import create_perturbation, PerturbationLike
from ...typing import Array1D


class FourierTransformer(BaseDistributor):

    """ Iteratively take the Fourier transform of density matrices.

    Parameters
    ----------
    rho_nn_reader
        Object that can iteratively read density matrices in the time domain,
        distributed such that different ranks have different chunks of the density
        matrix, but each ranks has all times for the same chunk.
    perturbation
        The perturbation which the density matrices are a response to.
    filter_frequencies
        After Fourier transformation keep only these frequencies (or the frequencies
        closest to them). In atomic units.
    frequency_broadening
        Gaussian broadening width in atomic units. Default (0) is no broadening.
    result_on_ranks
        List of ranks among which the resulting arrays will be distributed.
        Empty list (default) to distribute among all ranks.
    """

    def __init__(self,
                 rho_nn_reader: TimeDistributor,
                 perturbation: PerturbationLike,
                 filter_frequencies: list[float] | Array1D[np.float64] | None = None,
                 frequency_broadening: float = 0,
                 result_on_ranks: list[int] = []):
        super().__init__(rho_nn_reader.rho_wfs_reader,
                         rho_nn_reader._parameters,
                         comm=rho_nn_reader.comm)
        self.rho_nn_reader = rho_nn_reader
        self.perturbation = create_perturbation(perturbation)
        self.frequency_broadening = frequency_broadening
        self._flt_w = get_array_filter(self._omega_w, filter_frequencies)

        if len(result_on_ranks) == 0:
            self._result_on_ranks = set(range(self.comm.size))
        else:
            assert all(isinstance(rank, int) and rank >= 0 and rank < self.comm.size
                       for rank in result_on_ranks), result_on_ranks
            self._result_on_ranks = set(result_on_ranks)

        self._dist_buffer: DensityMatrixBuffer | None = None

    @property
    def dtype(self):
        return np.complex128

    @property
    def xshape(self):
        return (self.nw, )

    @property
    def freq_w(self) -> Array1D[np.float64]:
        return self._omega_w[self.flt_w]  # type: ignore

    @property
    def _omega_w(self) -> Array1D[np.float64]:
        padnt = fast_pad(self.rho_nn_reader.nt)
        dt = self.rho_nn_reader.dt
        omega_w = 2 * np.pi * np.fft.rfftfreq(padnt, dt)

        return omega_w  # type: ignore

    @property
    def nw(self) -> int:
        return len(self.freq_w)

    @property
    def nlocalw(self) -> int:
        return (self.nw + self.nranks_result - 1) // self.nranks_result

    @property
    def flt_w(self) -> slice | Array1D[np.bool_]:
        return self._flt_w

    @property
    def result_on_ranks(self) -> list[int]:
        """ Set of ranks among which the result will be distributed """
        return sorted(self._result_on_ranks)

    @property
    def nranks_result(self) -> int:
        """ Number of ranks that the resulting arrays will be distributed among """
        return len(self._result_on_ranks)

    def distributed_work(self) -> list[list[int]]:
        freqw_r = self.comm.size * [[]]
        for r, rank in enumerate(self.result_on_ranks):
            freqw_r[rank] = list(range(r, self.nw, self.nranks_result))

        return freqw_r

    def my_work(self) -> list[int]:
        freqw_r = self.distributed_work()
        return freqw_r[self.comm.rank]

    def __str__(self) -> str:
        nt = len(self.rho_nn_reader.time_t)
        niters = len(list(self.work_loop_by_ranks()))

        lines = []
        lines.append('Fourier transformer')
        lines.append(f'  Calculating Fourier transform on {self.maxnchunks} ranks')
        lines.append('  Fast Fourier transform')
        lines.append(f'    matrix dimensions {self.rho_nn_reader._parameters.nnshape}')
        lines.append(f'    grid of {nt} times')
        lines.append(f'    {self.describe_reim()}')
        if self.frequency_broadening == 0:
            lines.append('    No frequency broadening')
        else:
            lines.append(f'    Applying frequency broadening of {self.frequency_broadening * au_to_eV:.2f}eV')
        lines.append(f'    keeping frequency grid of {self.nw} frequencies')
        lines.append(f'      {format_frequencies(self.freq_w, units="au")}')
        lines.append('')

        lines.append('  Redistributing into full density matrices')
        lines.append(f'    {niters} iterations to process all chunks')
        lines.append(f'    matrix dimensions {self.rho_nn_reader._parameters.full_nnshape}')
        lines.append(f'    result stored on {self.nranks_result} ranks')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        parameters = self.rho_nn_reader._parameters

        narrays = 2 if self.yield_re and self.yield_im else 1
        temp_shape = parameters.nnshape + (self.maxnchunks, self.nlocalw, narrays)
        result_shape = parameters.full_nnshape + (self.nlocalw, narrays)

        total_result_size = int(np.prod(parameters.full_nnshape)) * self.nw * narrays

        comment = f'Buffers hold {narrays} arrays ({self.describe_reim()})'
        own_memory_estimate = MemoryEstimate(comment=comment)
        own_memory_estimate.add_key('Temporary buffer', temp_shape, complex,
                                    on_num_ranks=self.nranks_result)
        own_memory_estimate.add_key('Result buffer', result_shape, complex,
                                    total_size=total_result_size,
                                    on_num_ranks=self.nranks_result)

        memory_estimate = MemoryEstimate()
        memory_estimate.add_child('Time-dependent wave functions reader',
                                  self.rho_nn_reader.rho_wfs_reader.get_memory_estimate())
        memory_estimate.add_child('Parallel density matrices reader',
                                  self.rho_nn_reader.get_memory_estimate())
        memory_estimate.add_child('Fourier transformer',
                                  own_memory_estimate)

        return memory_estimate

    def __iter__(self) -> Generator[DensityMatrixBuffer, None, None]:
        time_t = self.rho_nn_reader.time_t  # Times in wave functions file
        dt = self.rho_nn_reader.dt  # Time step
        padnt = fast_pad(len(time_t))  # Pad with zeros

        dm_buffer = DensityMatrixBuffer(self.rho_nn_reader._parameters.nnshape,
                                        (self.nw, ),
                                        np.complex128)
        if self.yield_re:
            dm_buffer.zeros(True, 0)
        if self.yield_im:
            dm_buffer.zeros(False, 0)

        for read_buffer in self.rho_nn_reader:
            for data_nnt, buffer_nnw in zip(read_buffer._iter_buffers(), dm_buffer._iter_buffers()):
                if self.frequency_broadening == 0:
                    data_nnw = self.perturbation.normalize_frequency_response(data_nnt, time_t, padnt, axis=-1)
                else:
                    data_nnt = self.perturbation.normalize_time_response(data_nnt, time_t, axis=-1)
                    data_nnt[..., :len(time_t)] *= np.exp(-0.5 * self.frequency_broadening ** 2 * time_t**2)
                    data_nnw = np.fft.rfft(data_nnt, n=padnt, axis=-1) * dt
                buffer_nnw[:] = data_nnw[..., self.flt_w].conj()  # Change sign convention

            yield dm_buffer.copy()

    @property
    def dist_buffer(self) -> DensityMatrixBuffer:
        """ Buffer of density matrices on this rank after redistribution """
        if self._dist_buffer is None:
            self._dist_buffer = self.redistribute()
        return self._dist_buffer

    def create_out_buffer(self) -> DensityMatrixBuffer:
        """ Create the DensityMatrixBuffer to hold the temporary density matrix after each redistribution """
        parameters = self.rho_nn_reader._parameters
        nlocalw = self.nlocalw if self.comm.rank in self.result_on_ranks else 0
        out_dm = DensityMatrixBuffer(nnshape=parameters.nnshape,
                                     xshape=(self.maxnchunks, nlocalw),
                                     dtype=np.complex128)
        out_dm.zero_buffers(real=self.yield_re, imag=self.yield_im, derivative_order_s=[0])

        return out_dm

    def create_result_buffer(self) -> DensityMatrixBuffer:
        """ Create the DensityMatrixBuffer to hold the resulting density matrix """
        parameters = self.rho_nn_reader._parameters
        nnshape = parameters.full_nnshape
        full_dm = DensityMatrixBuffer(nnshape=nnshape,
                                      xshape=(len(self.my_work()), ),
                                      dtype=np.complex128)
        full_dm.zero_buffers(real=self.yield_re, imag=self.yield_im, derivative_order_s=[0])

        return full_dm

    def redistribute(self) -> DensityMatrixBuffer:
        """ Perform the Fourier transform and redistribute the data

        When the Fourier transform is performed, the data is distributed such that each rank
        stores the entire time/frequency series for one chunk of the density matrices, i.e. indices n1, n2.

        This function then performs a redistribution of the data such that each rank stores full
        density matrices, for certain frequencies.

        If the density matrices are split into more chunks than there are ranks, then the
        chunks are read, Fourier transformed and distributed in a loop several times until all
        data has been processed.

        Returns
        -------
        Density matrix buffer with x-dimensions (Number of local frequencies, )
        where the Number of local frequencies variers between the ranks.
        """
        local_work = iter(self)
        parameters = self.rho_nn_reader._parameters
        log = self.log
        self.rho_nn_reader.rho_wfs_reader.lcao_rho_reader.striden == 0, \
            'n stride must be 0 (index all) for redistribute'

        # Frequency indices of result on each rank
        freqw_r = self.distributed_work()
        niters = len(list(self.work_loop_by_ranks()))

        out_dm = self.create_out_buffer()
        full_dm = self.create_result_buffer()

        _exhausted = object()

        # Loop over the chunks of the density matrix
        for chunki, indices_r in enumerate(self.work_loop_by_ranks()):
            # At this point, each rank stores one unique chunk of the density matrix.
            # All ranks have the entire time series of data for their own chunk.
            # If there are more chunks than ranks, then this loop will run
            # for several iterations. If the number of chunks is not divisible by the number of
            # ranks then, during the last iteration, some of the chunks are None (meaning the rank
            # currently has no data).

            # List of chunks that each rank currently stores, where element r of the list
            # contains the chunk that rank r works with. Ranks higher than the length of the list
            # currently store no chunks.
            # The list itself is identical on all ranks.
            chunks_by_rank = [indices[2:] for indices in indices_r if indices is not None]

            ntargets = len(chunks_by_rank)

            if self.comm.rank < ntargets:
                # This rank has data to send. Compute the Fourier transform and store the result
                dm_buffer = next(local_work)
            else:
                # This rank has no data to send
                assert next(local_work, _exhausted) is _exhausted
                # Still, we need to create a dummy buffer
                dm_buffer = DensityMatrixBuffer(nnshape=parameters.nnshape,
                                                xshape=(0, ), dtype=np.complex128)
                dm_buffer.zero_buffers(real=self.yield_re, imag=self.yield_im,
                                       derivative_order_s=[0])

            log.start('alltoallv')

            # Redistribute the data:
            # - dm_buffer stores single chunks of density matrices, for all frequencies.
            # - out_dm will store several chunks, for a few frequencies.
            # source_indices_r describes which slices of dm_buffer should be sent to which rank
            # target_indices_r describes to which positions of the out_dm buffer should be received
            # from which rank
            source_indices_r = [None if len(w) == 0 else w for w in freqw_r]
            target_indices_r = [r if r < ntargets else None for r in range(self.comm.size)]
            dm_buffer.redistribute(out_dm,
                                   comm=self.comm,
                                   source_indices_r=source_indices_r,
                                   target_indices_r=target_indices_r,
                                   log=log)

            if self.comm.rank == 0:
                log(f'Chunk {chunki+1}/{niters}: distributed frequency response in '
                    f'{log.elapsed("alltoallv"):.1f}s', flush=True, who='Response')

            for array_nnrw, full_array_nnw in zip(out_dm._iter_buffers(), full_dm._iter_buffers()):
                for r, nn_indices in enumerate(chunks_by_rank):
                    safe_fill_larger(full_array_nnw[nn_indices], array_nnrw[:, :, r])

        assert next(local_work, _exhausted) is _exhausted

        return full_dm

    @classmethod
    def from_reader(cls,  # type: ignore
                    rho_nn_reader: KohnShamRhoWfsReader,
                    parameters: RhoParameters,
                    *,
                    perturbation: PerturbationLike,
                    filter_frequencies: list[float] | Array1D[np.float64] | None = None,
                    frequency_broadening: float = 0,
                    result_on_ranks: list[int] = []) -> FourierTransformer:
        time_distributor = AlltoallvTimeDistributor(rho_nn_reader, parameters)
        fourier_transformer = FourierTransformer(time_distributor,
                                                 perturbation=perturbation,
                                                 filter_frequencies=filter_frequencies,
                                                 frequency_broadening=frequency_broadening,
                                                 result_on_ranks=result_on_ranks)
        return fourier_transformer



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generator, Collection

import numpy as np
from numpy.typing import NDArray

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.mpi import world

from ..typing import Communicator, Array2D, Array3D
from ..response import BaseResponse
from ..perturbation import create_perturbation, Perturbation, PerturbationLike
from ..density_matrices.base import BaseDensityMatrices, WorkMetadata
from ..density_matrices.time import ConvolutionDensityMatrices
from ..density_matrices.frequency import FrequencyDensityMatrices
from ..voronoi import VoronoiWeights, EmptyVoronoiWeights
from ..typing import Array1D
from ..utils import Logger, ResultKeys, Result, broaden_n2e, broaden_xn2e, broaden_ia2ou
from ..utils.memory import HasMemoryEstimate, MemoryEstimate


class BaseObservableCalculator(HasMemoryEstimate, ABC):

    """ Object of this class compute observables.

    Parameters
    ----------
    response
        Response object.
    voronoi
        Voronoi weights object.
    energies_occ
        Energy grid in units of eV for occupied levels (holes).
    energies_unocc
        Energy grid in units of eV for unoccupied levels (electrons).
    sigma
        Gaussian broadening width for energy grid in units of eV.
    times
        Compute observables in the time domain, for these times (or as close to them as possible).
        In units of as.

        May not be used together with :attr:`frequencies` or :attr:`frequency_broadening`.
    pulses
        Compute observables in the time domain, in response to these pulses.
        If none, then no pulse convolution is performed.

        May not be used together with :attr:`frequencies` or :attr:`frequency_broadening`.
    frequencies
        Compute observables in the frequency domain, for these frequencies. In units of eV.

        May not be used together with :attr:`times` or :attr:`pulses`.
    frequency_broadening
        Compute observables in the frequency domain, with Gaussian broadening of this width.
        In units of eV.

        May not be used together with :attr:`times` or :attr:`pulses`.
    """

    def __init__(self,
                 response: BaseResponse,
                 voronoi: VoronoiWeights | None = None,
                 *,
                 energies_occ: list[float] | Array1D[np.float64] | None = None,
                 energies_unocc: list[float] | Array1D[np.float64] | None = None,
                 sigma: float | None = None,
                 times: list[float] | Array1D[np.float64] | None = None,
                 pulses: Collection[PerturbationLike] | None = None,
                 frequencies: list[float] | Array1D[np.float64] | None = None,
                 frequency_broadening: float = 0,
                 ):
        self._log = Logger()
        self.log.startup_message()
        world.barrier()
        self._response = response

        self._is_time_density_matrices: bool
        if times is not None:
            if frequencies is not None or frequency_broadening != 0:
                raise ValueError('The parameters frequencies and frequency_broadening may not '
                                 'be used together with times and pulses.')
            # Time calculator
            if pulses is None:
                pulses = [None]  # No perturbation
            pulses = [create_perturbation(pulse) for pulse in pulses]
            self._is_time_density_matrices = True
        else:
            if frequencies is None:
                raise ValueError('One of the parameters times or frequencies must be given.')
            if pulses is not None:
                raise ValueError('The parameters frequencies and frequency_broadening may not '
                                 'be used together with times and pulses.')
            # Frequency calculator
            self._is_time_density_matrices = False

        derivatives, real, imag = self._need_derivatives_real_imag

        density_matrices: BaseDensityMatrices
        if self._is_time_density_matrices:
            assert times is not None
            assert pulses is not None
            density_matrices = response._get_time_density_matrices(
                    times, pulses, derivatives, real, imag, self.log)
        else:
            assert pulses is None
            assert frequencies is not None
            density_matrices = response._get_frequency_density_matrices(
                    frequencies, frequency_broadening, real, imag, self.log)

        self._density_matrices = density_matrices
        self._voronoi: VoronoiWeights
        if voronoi is None:
            self._voronoi = EmptyVoronoiWeights()
        else:
            self._voronoi = voronoi
        if energies_occ is None:
            energies_occ = []
        if energies_unocc is None:
            energies_unocc = []
        self._energies_occ = np.asarray(energies_occ)
        self._energies_unocc = np.asarray(energies_unocc)
        self._sigma = sigma
        self._weight_In: Array2D[np.float64] | None = None
        self._weight_Iii: Array3D[np.float64] | None = None
        self._weight_Iaa: Array3D[np.float64] | None = None
        world.barrier()
        self.log(f'Set up calculator:\n'
                 f'{self}\n\n'
                 '==================================\n'
                 ' Procedure for obtaining response \n'
                 '==================================\n'
                 f'{self.density_matrices}\n\n'
                 '=================\n'
                 ' Memory estimate \n'
                 '=================\n'
                 f'{self.get_memory_estimate()}\n',
                 rank=0)

    def __str__(self) -> str:
        lines = [f' {self.__class__.__name__} ']

        lines.append('response:')
        lines += ['  ' + line for line in str(self.response).split('\n')]
        lines.append('')

        lines.append('voronoi:')
        if self.nproj == 0:
            lines.append('  No Voronoi decomposition')
        else:
            lines += ['  ' + line for line in str(self.voronoi).split('\n')]
        lines.append('')

        if len(self.energies_occ) == 0 or len(self.energies_unocc) == 0 or self.sigma is None:
            lines.append('No energies for broadening')
        else:
            lines += [f'Energies for broadening (sigma = {self.sigma:.1f} eV)',
                      f'  Occupied: {len(self.energies_occ)} from '
                      f'{self.energies_occ[0]:.1f} to {self.energies_occ[-1]:.1f} eV',
                      f'  Unoccupied: {len(self.energies_unocc)} from '
                      f'{self.energies_unocc[0]:.1f} to {self.energies_unocc[-1]:.1f} eV',
                      ]

        # Make a cute frame
        maxlen = max(len(line) for line in lines)
        lines[0] = '{s:=^{n}}'.format(s=f' {lines[0]} ', n=maxlen)
        lines.append(maxlen * '-')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        memory_estimate = MemoryEstimate()
        memory_estimate.add_child('Reading response', self.density_matrices.get_memory_estimate())
        for key, shape in self._voronoi_shapes.items():
            memory_estimate.add_key(f'Voronoi weights {key}', shape, float,
                                    on_num_ranks=self.loop_comm.size)

        return memory_estimate

    @property
    def response(self) -> BaseResponse:
        """ Response object """
        return self._response

    @property
    def density_matrices(self) -> BaseDensityMatrices:
        """ Object that gives the density matrix in the time or freqency domain. """
        return self._density_matrices

    @property
    def nproj(self) -> int:
        """ Number of projections in the Voronoi weights object """
        return self.voronoi.nproj

    @property
    def voronoi(self) -> VoronoiWeights:
        """ Voronoi weights object """
        return self._voronoi

    @property
    def energies_occ(self) -> NDArray[np.float64]:
        """ Energy grid (in units of eV) for occupied levels (hot holes). """
        return self._energies_occ

    @property
    def energies_unocc(self) -> NDArray[np.float64]:
        """ Energy grid (in units of eV) for unoccupied levels (hot electrons). """
        return self._energies_unocc

    @property
    def sigma(self) -> float | None:
        """ Gaussian broadening width in units of eV. """
        return self._sigma

    @property
    def ksd(self) -> KohnShamDecomposition:
        """ Kohn-Sham decomposition object """
        return self.density_matrices.ksd

    @property
    def log(self) -> Logger:
        """ Logger """
        return self._log

    def log_parallel(self, *args, **kwargs) -> Logger:
        """ Log message with communicator information. """
        return self._log(*args, **kwargs, comm=self.loop_comm, who='Calculator')

    @property
    def frequencies(self) -> Array1D[np.float64]:
        """ Frequencies (in units of eV) at which the density matrices are evaluated.

        Only valid when the density matrices object is defined in the frequency domain. """
        assert isinstance(self.density_matrices, FrequencyDensityMatrices)
        return self.density_matrices.frequencies

    @property
    def frequency_broadening(self) -> float:
        """ Value of frequency broadening in units of eV.

        Only valid when the density matrices object is defined in the frequency domain. """
        assert isinstance(self.density_matrices, FrequencyDensityMatrices)
        return self.density_matrices.frequency_broadening

    @property
    def times(self) -> Array1D[np.float64]:
        """ Times (in units of as) at which the density matrices are evaluated.

        Only valid when the density matrices object is defined in the time domain. """
        assert isinstance(self.density_matrices, ConvolutionDensityMatrices)
        return self.density_matrices.times

    @property
    def pulses(self) -> list[Perturbation]:
        """ List of pulses which the density matrices are responses to.

        Only valid when the density matrices object is a convolution with pulses. """
        assert isinstance(self.density_matrices, ConvolutionDensityMatrices)
        return self.density_matrices.pulses

    @property
    def calc_comm(self) -> Communicator:
        """ Calculation communicator.

        Each rank of this communicator calculates the observables corresponding to
        a part (in electron-hole space) of the density matrices. """
        return self.density_matrices.calc_comm

    @property
    def loop_comm(self) -> Communicator:
        """ Loop communicator.

        Each rank of this communicator calculates the density matrices corresponding to
        different times, frequencies or after convolution with a different pulse. """
        return self.density_matrices.loop_comm

    @property
    def eig_n(self) -> Array1D[np.float64]:
        """ Eigenvalues (in units of eV) relative to Fermi level in the full ground state KS basis. """
        eig_n, _ = self.ksd.get_eig_n(zero_fermilevel=True)
        return eig_n  # type: ignore

    @property
    def eig_i(self) -> Array1D[np.float64]:
        """ Eigenvalues (in units of eV) relative to Fermi level of occupied states (holes). """
        return self.eig_n[self.flti]  # type: ignore

    @property
    def eig_a(self) -> Array1D[np.float64]:
        """ Eigenvalues (in units of eV) relative to Fermi level of unoccupied states (electrons). """
        return self.eig_n[self.flta]  # type: ignore

    @property
    def flti(self) -> slice:
        """ Slice for extracting indices corresponding to occupied states. """
        imin, imax, _, _ = self.ksd.ialims()
        return slice(imin, imax+1)

    @property
    def flta(self) -> slice:
        """ Slice for extracting indices corresponding to unoccupied states. """
        _, _, amin, amax = self.ksd.ialims()
        return slice(amin, amax+1)

    @property
    def _need_derivatives_real_imag(self) -> tuple[list[int], bool, bool]:
        """ Derivatives needed by this calculator, and
        whether real and imaginary parts are needed
        """
        raise NotImplementedError

    @property
    def _voronoi_shapes(self) -> dict[str, tuple[int, ...]]:
        """ List of shapes for the Voronoi weights stored on this calculator. """
        return {}

    def _read_weights_diagonal(self) -> None:
        """ Read the diagonal weights from the voronoi object and store them in memory. """
        nI = self.voronoi.nproj
        if nI == 0:
            return
        Nn = self.voronoi.nn

        if self.calc_comm.rank == 0:
            weight_In = np.empty((nI, Nn), dtype=float)
        else:
            weight_In = None
        for iI, weight_nn in enumerate(self.voronoi):
            if self.voronoi.comm.rank == 0:
                assert weight_In is not None
                assert weight_nn is not None
                weight_In[iI, ...] = weight_nn.diagonal()
            else:
                assert weight_nn is None

        if self.calc_comm.rank == 0:
            # Broadcast to all calc_comm rank 0's
            self.loop_comm.broadcast(weight_In, 0)

        self._weight_In = weight_In

    def _read_weights_eh(self) -> None:
        """ Read the electron-hole weights from the voronoi object and store them in memory. """
        nI = self.voronoi.nproj
        if nI == 0:
            return

        if self.calc_comm.rank == 0:
            weight_Iii = np.empty((nI, len(self.eig_i), len(self.eig_i)), dtype=float)
            weight_Iaa = np.empty((nI, len(self.eig_a), len(self.eig_a)), dtype=float)
        else:
            weight_Iii = None
            weight_Iaa = None
        for iI, weight_nn in enumerate(self.voronoi):
            if self.voronoi.comm.rank == 0:
                assert weight_nn is not None
                assert weight_Iii is not None
                assert weight_Iaa is not None
                weight_Iii[iI, ...] = weight_nn[self.flti, self.flti]
                weight_Iaa[iI, ...] = weight_nn[self.flta, self.flta]
            else:
                assert weight_nn is None

        if self.calc_comm.rank == 0:
            # Broadcast to all calc_comm rank 0's
            self.loop_comm.broadcast(weight_Iii, 0)
            self.loop_comm.broadcast(weight_Iaa, 0)

        self._weight_Iii = weight_Iii
        self._weight_Iaa = weight_Iaa

    @property
    def _iterate_weights_diagonal(self) -> Generator[NDArray[np.float64], None, None]:
        """ Iterate over the diagonal weights.

        Yields
        ------
        The diagonal of the Voronoi weights, one projection at a time """
        assert self.calc_comm.rank == 0
        if self.nproj == 0:
            return

        if self._weight_In is None:
            self._read_weights_diagonal()
        assert self._weight_In is not None

        for weight_n in self._weight_In:
            yield weight_n

    @property
    def _iterate_weights_eh(self) -> Generator[tuple[NDArray[np.float64], NDArray[np.float64]], None, None]:
        """ Iterate over the electron-hole part of the weights.

        Yields
        ------
        The electron-hole part of the Voronoi weights, one projection at a time """
        assert self.calc_comm.rank == 0
        if self.nproj == 0:
            return

        if self._weight_Iaa is None or self._weight_Iii is None:
            self._read_weights_eh()
        assert self._weight_Iii is not None
        assert self._weight_Iaa is not None

        for weight_ii, weight_aa in zip(self._weight_Iii, self._weight_Iaa):
            yield weight_ii, weight_aa

    @abstractmethod
    def get_result_keys(self) -> ResultKeys:
        """ Get the keys that each result will contain, and dimensions thereof.

        Returns
        -------
        Object representing the data that will be present in the result objects. """
        raise NotImplementedError

    @abstractmethod
    def icalculate(self) -> Generator[tuple[WorkMetadata, Result], None, None]:
        """ Iteratively calculate results.

        Yields
        ------
        Tuple (work, result) on the root rank of the calculation communicator:

        work
            An object representing the metadata (time, frequency or pulse) for the work done
        result
            Object containg the calculation results for this time, frequency or pulse

        Yields nothing on non-root ranks of the calculation communicator.
        """
        raise NotImplementedError

    def icalculate_gather_on_root(self, **kwargs) -> Generator[
            tuple[WorkMetadata, Result], None, None]:
        """ Iteratively calculate results and gather to the root rank.

        Yields
        ------
        Tuple (work, result) on the root rank of the both calculation and loop communicators:

        work
            An object representing the metadata (time, frequency or pulse) for the work done
        result
            Object containg the calculation results for this time, frequency or pulse

        Yields nothing on non-root ranks of the communicators.
        """
        resultkeys = self.get_result_keys(**kwargs)
        gen = iter(self.icalculate(**kwargs))

        # Loop over the work to be done, and the ranks that are supposed to do it
        for rank, work in self.density_matrices.global_work_loop_with_idle():
            if work is None:
                # Rank rank will not do any work at this point
                continue

            if self.calc_comm.rank > 0:
                continue

            if self.loop_comm.rank == 0 and rank == 0:
                # This is the root rank, and the root rank should yield its own work
                mywork, result = next(gen)
                assert work.global_indices == mywork.global_indices, f'{work.desc} != {mywork.desc}'
                yield mywork, result
                # self.log.start('communicate')
            elif self.loop_comm.rank == 0:
                # This is the root rank, and the root rank should receive the work
                # done by rank rank, and yield that
                result.inplace_receive(resultkeys, rank, comm=self.loop_comm)
                yield work, result
                # self.log(f'Communicated for {self.log.elapsed("communicate"):.2f}s', flush=True)
            elif self.loop_comm.rank == rank:
                # This is not the root rank, but this rank should send its data to root
                _, result = next(gen)
                result.send(resultkeys, 0, comm=self.loop_comm)

        _exhausted = object()
        rem = next(gen, _exhausted)
        assert rem is _exhausted, rem

    def write_response_to_disk(self,
                               fmt: str):
        """ Calculate the response needed for this calculator and write the response to disk.

        The necessary derivatives and real and imaginary parts of the density matrices
        will be written. Can be used both in the time and frequency domains.

        Parameters
        ----------
        fmt
            Formatting string for the density matrices saved to disk.

            The formatting string should be a plain string containing variable
            placeholders within curly brackets `{}`. It should not be confused with
            a formatted string literal (f-string).

            Examples:

            * frho_fmt = `'frho/w{freq:05.2f}-{reim}.npy'`.
            * pulserho_fmt =  `pulserho/t{time:09.1f}{tag}.npy`.

            Accepts variables in the time domain:

            * `{time}` - Time in units of as.
            * `{tag}` - Derivative tag, `''`, `'-Iomega'`, or `'-omega2'`.
            * `{pulsefreq}` - Pulse frequency in units of eV.
            * `{pulsefwhm}` - Pulse FWHM in units of fs.

            Accepts variables in the frequency domain:

            * `{freq}` - Frequency in units of eV.
            * `{reim}` - `'Re'` or `'Im'` for Fourier transform of real/imaginary
               part of density matrix.
        """
        self._density_matrices.write_to_disk(fmt)

    def broaden_occ(self,
                    M_i: NDArray[np.float64]) -> NDArray[np.float64]:
        """ Broaden array corresponding to occupied levels with Gaussians of width sigma

        Parameters
        ----------
        M_i
            Array to broaden

        Returns
        -------
        Broadened array
        """
        assert self.sigma is not None

        return broaden_n2e(M_i, self.eig_i, self.energies_occ, self.sigma)

    def broaden_unocc(self,
                      M_a: NDArray[np.float64]) -> NDArray[np.float64]:
        """ Broaden array corresponding to unoccupied levels with Gaussians of width sigma

        Parameters
        ----------
        M_a
            Array to broaden

        Returns
        -------
            Broadened array
        """
        assert self.sigma is not None

        return broaden_n2e(M_a, self.eig_a, self.energies_unocc, self.sigma)

    def broaden_xi2o(self,
                     M_xi: NDArray[np.float64]) -> NDArray[np.float64]:
        """ Broaden array corresponding to occupied levels with Gaussians of width sigma

        Parameters
        ----------
        M_xa
            Array to broaden. The last dimension should correspond to the occupied levels

        Returns
        -------
        Broadened array
        """
        assert self.sigma is not None

        return broaden_xn2e(M_xi, self.eig_i, self.energies_occ, self.sigma)

    def broaden_xi2u(self,
                     M_xa: NDArray[np.float64]) -> NDArray[np.float64]:
        """ Broaden array corresponding to unoccupied levels with Gaussians of width sigma

        Parameters
        ----------
        M_xa
            Array to broaden. The last dimension should correspond to the unoccupied levels

        Returns
        -------
            Broadened array
        """
        assert self.sigma is not None

        return broaden_xn2e(M_xa, self.eig_a, self.energies_unocc, self.sigma)

    def broaden_ia2ou(self,
                      M_ia: NDArray[np.float64]) -> NDArray[np.float64]:
        """ Broaden matrix in electron-hole basis with Gaussians of width sigma.

        Parameters
        ----------
        M_ia
            Matrix to broaden. The first dimension should correspond to occupied levels
            and the second to unoccupied levels.

        Returns
        -------
            Broadened array
        """
        assert self.sigma is not None

        return broaden_ia2ou(M_ia, self.eig_i, self.eig_a,
                             self.energies_occ, self.energies_unocc, self.sigma)



from __future__ import annotations

import numpy as np
from typing import Any, Generator
from numpy.typing import NDArray

from gpaw.tddft.units import au_to_eA, au_to_eV

from .base import BaseObservableCalculator
from ..density_matrices.base import WorkMetadata
from ..utils import ResultKeys, Result


class DipoleCalculator(BaseObservableCalculator):

    r""" Calculate the induced dipole moment in the time or frequency domain.

    The induced dipole moment (i.e. the dipole moment minus the permanent
    component) is to first order given by

    .. math::

        \delta\boldsymbol{\mu} = -2 \sum_{ia}^\text{eh}
        \boldsymbol{\mu}_{ia} \mathrm{Re}\:\delta\rho_{ia},

    where :math:`\boldsymbol{\mu}_{ia}` is the dipole matrix element of ground state Kohn-Sham
    pair :math:`ia`

    .. math::

        \boldsymbol{\mu}_{ia} = \int \psi^{(0)}_i(\boldsymbol{r}) \boldsymbol{r}
        \psi^{(0)}_a(\boldsymbol{r}) \mathrm{d}\boldsymbol{r},

    and :math:`\delta\rho_{ia}` the induced Kohn-Sham density matrix.

    In the frequency domain, this calculator calculates the polarizability, i.e. the Fourier
    transform of the dipole moment divided by the perturbation.

    .. math::

        \boldsymbol{\alpha}(\omega) = -2 \sum_{ia}^\text{eh}
        \boldsymbol{\mu}_{ia} \frac{\mathcal{F}\left[\mathrm{Re}\:\delta\rho_{ia}\right](\omega)}{v(\omega)}.

    The absorption spectrum in units of dipole strength function is the imaginary part
    of the polarizability times a prefactor

    .. math::

        \boldsymbol{S}(\omega) = \frac{2\omega}{\pi} \mathrm{Im}\:\boldsymbol{\alpha}(\omega).

    This class can also compute projections of the above on Voronoi weights :math:`w_{ia}`.

    Parameters
    ----------
    response
        Response object.
    voronoi
        Voronoi weights object.
    energies_occ
        Energy grid in units of eV for occupied levels (holes).
    energies_unocc
        Energy grid in units of eV for unoccupied levels (electrons)
    sigma
        Gaussian broadening width for energy grid in units of eV.
    times
        Compute induced dipole in the time domain, for these times (or as close to them as possible).
        In units of as.

        May not be used together with :attr:`frequencies` or :attr:`frequency_broadening`.
    pulses
        Compute induced dipole in the time domain, in response to these pulses.
        If none, then no pulse convolution is performed.

        May not be used together with :attr:`frequencies` or :attr:`frequency_broadening`.
    frequencies
        Compute polarizability in the frequency domain, for these frequencies. In units of eV.

        May not be used together with :attr:`times` or :attr:`pulses`.
    frequency_broadening
        Compute polarizability in the frequency domain, with Gaussian broadening of this width.
        In units of eV.

        May not be used together with :attr:`times` or :attr:`pulses`.
    """

    def get_result_keys(self,
                        yield_total_ia: bool = False,
                        yield_proj_ia: bool = False,
                        yield_total_ou: bool = False,
                        yield_proj_ou: bool = False,
                        decompose_v: bool = True,
                        v: int | None = None,
                        ) -> ResultKeys:
        r""" Get the keys that each result will contain, and dimensions thereof.

        Parameters
        ----------
        yield_total_ia
            The results should include the total dipole contributions in the electron-hole basis
            :math:`-2 \boldsymbol{\mu}_{ia} \mathrm{Re}\:\delta\rho_{ia}`.
        yield_proj_ia
            The results should include projections of the dipole contributions in the electron-hole basis
            :math:`-2 \boldsymbol{\mu}_{ia} \mathrm{Re}\:\delta\rho_{ia} w_{ia}`.
        yield_total_ou
            The results should include the total dipole contributions on the energy grid.
        yield_proj_ou
            The results should include projections of the dipole contributions on the energy grid.
        decompose_v
            The results should include the dipole moment and/or its contributions decomposed
            by Cartesian direction.
        v
            If not None, then the results should include the v:th Cartesian component
            of the dipole moment and its contributions.
        """
        assert decompose_v or v is not None

        nI = self.nproj
        imin, imax, amin, amax = self.ksd.ialims()
        ni = imax - imin + 1
        na = amax - amin + 1
        no = len(self.energies_occ)
        nu = len(self.energies_unocc)

        # Time domain: save dipole moment, which is real
        # Frequency domain: save polarizability, which is complex
        dtype = float if self._is_time_density_matrices else complex

        resultkeys = ResultKeys()
        if v is not None:
            resultkeys.add_key('dm', dtype=dtype)
            if yield_total_ia:
                resultkeys.add_key('dm_ia', (ni, na), dtype=dtype)
            if yield_total_ou:
                resultkeys.add_key('dm_ou', (no, nu), dtype=dtype)

            resultkeys.add_key('dm_proj_II', (nI, nI), dtype=dtype)
            if yield_proj_ia:
                resultkeys.add_key('dm_occ_proj_Iia', (nI, ni, na), dtype=dtype)
                resultkeys.add_key('dm_unocc_proj_Iia', (nI, ni, na), dtype=dtype)
            if yield_proj_ou:
                resultkeys.add_key('dm_occ_proj_Iou', (nI, no, nu), dtype=dtype)
                resultkeys.add_key('dm_unocc_proj_Iou', (nI, no, nu), dtype=dtype)
        if decompose_v:
            resultkeys.add_key('dm_v', 3, dtype=dtype)
            if yield_total_ia:
                resultkeys.add_key('dm_iav', (ni, na, 3), dtype=dtype)
            if yield_total_ou:
                resultkeys.add_key('dm_ouv', (no, nu, 3), dtype=dtype)

            resultkeys.add_key('dm_proj_IIv', (nI, nI, 3), dtype=dtype)
            if yield_proj_ia:
                resultkeys.add_key('dm_occ_proj_Iiav', (nI, ni, na, 3), dtype=dtype)
                resultkeys.add_key('dm_unocc_proj_Iiav', (nI, ni, na, 3), dtype=dtype)
            if yield_proj_ou:
                resultkeys.add_key('dm_occ_proj_Iouv', (nI, no, nu, 3), dtype=dtype)
                resultkeys.add_key('dm_unocc_proj_Iouv', (nI, no, nu, 3), dtype=dtype)

        return resultkeys

    def icalculate(self,
                   yield_total_ia: bool = False,
                   yield_proj_ia: bool = False,
                   yield_total_ou: bool = False,
                   yield_proj_ou: bool = False,
                   decompose_v: bool = True,
                   v: int | None = None,
                   ) -> Generator[tuple[WorkMetadata, Result], None, None]:
        """ Iteratively calculate dipole contributions.

        Parameters
        ----------
        yield_total_ia
            The results should include the total dipole contributions in the electron-hole basis
            :math:`-2 \\boldsymbol{\\mu}_{ia} \\mathrm{Re}\\:\\delta\\rho_{ia}`.
        yield_proj_ia
            The results should include projections of the dipole contributions in the electron-hole basis
            :math:`-2 \\boldsymbol{\\mu}_{ia} \\mathrm{Re}\\:\\delta\\rho_{ia} w_{ia}`.
        yield_total_ou
            The results should include the total dipole contributions on the energy grid.
        yield_proj_ou
            The results should include projections of the dipole contributions on the energy grid.
        decompose_v
            The results should include the dipole moment and/or its contributions decomposed
            by Cartesian direction.
        v
            If not `None`, then the results should include the v:th Cartesian component \
            of the dipole moment and its contributions.

        Yields
        ------
        Tuple (work, result) on the root rank of the calculation communicator. \
        Does not yield on non-root ranks of the calculation communicator.

            work
                An object representing the metadata (time, frequency or pulse) for the work done.
            result
                Object containg the calculation results for this time, frequency or pulse.
        """
        include_energy_dists = (yield_total_ou or yield_proj_ou)
        if include_energy_dists:
            assert self.sigma is not None
        need_entire_matrix = (yield_total_ou or yield_proj_ou
                              or yield_total_ia or yield_proj_ia
                              or self.nproj > 0)

        # Time domain: dipole moment in units of eÅ
        # Frequency domain: polarizability in units of (eÅ)**2/eV
        unit = au_to_eA if self._is_time_density_matrices else au_to_eA**2 / au_to_eV
        dm0_iav = -2 * np.moveaxis(np.array([self.ksd.M_ia_from_M_p(dm_p) for dm_p in self.ksd.dm_vp]), 0, -1) * unit

        # List all keys that this method computes
        # This is necessary to safely send and receive data across ranks
        resultkeys = self.get_result_keys(yield_total_ia=yield_total_ia,
                                          yield_proj_ia=yield_proj_ia,
                                          yield_total_ou=yield_total_ou,
                                          yield_proj_ou=yield_proj_ou,
                                          decompose_v=decompose_v,
                                          v=v,
                                          )

        self._read_weights_diagonal()

        # Iterate over the pulses and times
        for work, dm in self.density_matrices:
            if dm.rank > 0:
                continue

            self.log.start('calculate')

            if need_entire_matrix:
                if self._is_time_density_matrices:
                    # Real dipole
                    dm_iav = dm0_iav * dm.rho_ia[..., None].real
                else:
                    # Complex polarizability
                    dm_iav = dm0_iav * dm.rho_ia[..., None]
                dm_v = np.sum(dm_iav, axis=(0, 1))

                if yield_total_ou:
                    dm_ouv = self.broaden_ia2ou(dm_iav)
            else:
                if self._is_time_density_matrices:
                    # Real dipole
                    dm_v = np.einsum('iav,ia->v', dm0_iav, dm.rho_ia.real, optimize=True)
                else:
                    # Complex polarizability
                    dm_v = np.einsum('iav,ia->v', dm0_iav, dm.rho_ia, optimize=True)

            result = Result()
            if decompose_v:
                result['dm_v'] = dm_v
                if yield_total_ia:
                    result['dm_iav'] = dm_iav
                if yield_total_ou:
                    result['dm_ouv'] = dm_ouv
            if v is not None:
                result['dm'] = dm_v[v]
                if yield_total_ia:
                    result['dm_ia'] = dm_iav[..., v]
                if yield_total_ou:
                    result['dm_ou'] = dm_ouv[..., v]

            # Initialize the remaining empty arrays
            result.create_all_empty(resultkeys)

            # Iterate over projections
            for iI, weight_n in enumerate(self._iterate_weights_diagonal):
                assert weight_n is not None
                weight_i = weight_n[self.flti]
                weight_a = weight_n[self.flta]

                for iI2, weight2_n in enumerate(self._iterate_weights_diagonal):
                    assert weight2_n is not None
                    weight2_a = weight2_n[self.flta]

                    dm_proj_v = np.einsum('iav,i,a->v', dm_iav, weight_i, weight2_a, optimize=True)
                    result['dm_proj_IIv'][iI, iI2] = dm_proj_v

                if yield_proj_ia or yield_proj_ou:
                    dm_occ_proj_iav = dm_iav * weight_i[:, None, None]
                    dm_unocc_proj_iav = dm_iav * weight_a[None, :, None]

                    if yield_proj_ou:
                        dm_occ_proj_ouv = self.broaden_ia2ou(dm_occ_proj_iav)
                        dm_unocc_proj_ouv = self.broaden_ia2ou(dm_unocc_proj_iav)

                    if decompose_v:
                        if yield_proj_ia:
                            result['dm_occ_proj_Iiav'][iI] = dm_occ_proj_iav
                            result['dm_unocc_proj_Iiav'][iI] = dm_unocc_proj_iav
                        if yield_proj_ou:
                            result['dm_occ_proj_Iouv'][iI] = dm_occ_proj_ouv
                            result['dm_unocc_proj_Iouv'][iI] = dm_unocc_proj_ouv
                    if v is not None:
                        if yield_proj_ia:
                            result['dm_occ_proj_Iia'][iI] = dm_occ_proj_iav[..., v]
                            result['dm_unocc_proj_Iia'][iI] = dm_unocc_proj_iav[..., v]
                        if yield_proj_ou:
                            dm_occ_proj_ouv = self.broaden_ia2ou(dm_occ_proj_iav)
                            dm_unocc_proj_ouv = self.broaden_ia2ou(dm_unocc_proj_iav)

                        if decompose_v:
                            if yield_proj_ia:
                                result['dm_occ_proj_Iiav'][iI] = dm_occ_proj_iav
                                result['dm_unocc_proj_Iiav'][iI] = dm_unocc_proj_iav
                            if yield_proj_ou:
                                result['dm_occ_proj_Iouv'][iI] = dm_occ_proj_ouv
                                result['dm_unocc_proj_Iouv'][iI] = dm_unocc_proj_ouv
                        if v is not None:
                            if yield_proj_ia:
                                result['dm_occ_proj_Iia'][iI] = dm_occ_proj_iav[..., v]
                                result['dm_unocc_proj_Iia'][iI] = dm_unocc_proj_iav[..., v]
                            if yield_proj_ou:
                                result['dm_occ_proj_Iou'][iI] = dm_occ_proj_ouv[..., v]
                                result['dm_unocc_proj_Iou'][iI] = dm_unocc_proj_ouv[..., v]

            self.log_parallel(f'Calculated and broadened dipoles contributions in {self.log.elapsed("calculate"):.2f}s '
                              f'for {work.desc}', flush=True)

            yield work, result

        if self.calc_comm.rank == 0:
            self.log_parallel('Finished calculating dipoles contributions', flush=True)

    @property
    def _need_derivatives_real_imag(self) -> tuple[list[int], bool, bool]:
        # Time domain: We only need the real part of the density matrix.
        # Frequency domain: We need the (complex) Fourier transform of
        #                   the real part of the density matrix.
        return ([0], True, False)

    @property
    def _voronoi_shapes(self) -> dict[str, tuple[int, ...]]:
        nI = self.voronoi.nproj
        if nI == 0:
            return {}
        Nn = self.voronoi.nn
        # Diagonal weights only
        return {'diagonal': (nI, Nn)}

    @property
    def oscillator_strength_prefactor(self) -> NDArray[np.float64]:
        """ Conversion factor from polarizability to dipole strength function.

        """
        from gpaw.tddft.units import eA_to_au, eV_to_au
        # Convert polarizability (eÅ**2/eV) to atomic units
        # Multiply by 2 omega / pi in atomic units
        # Convert to units of dipole strength function (1/eV)
        prefactor_w = 2 * (self.frequencies * eV_to_au) / np.pi * eA_to_au ** 2
        return prefactor_w

    def calculate_and_write(self,
                            out_fname: str,
                            write_extra: dict[str, Any] = dict(),
                            include_tcm: bool = False,
                            only_one_pulse: bool = True):
        """ Calculate induced dipole moments and transition contribution maps.

        Dipole moments and contributions are saved in a numpy archive if
        the file extension is `.npz` or in an ULM file if the file extension is `.ulm`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of extra key-value pairs to write to the data file.
        include_tcm
            Whether the transition contribution map (TCM) to the dipole moments should be computed and saved.
        only_one_pulse
            If False, group arrays by pulse. Only valid in time domain.
        """
        from ..writers.tcm import DipoleWriter
        from ..writers.writer import FrequencyResultsCollector, TimeResultsCollector, PulseConvolutionResultsCollector

        cls = ((TimeResultsCollector if only_one_pulse else PulseConvolutionResultsCollector)
               if self._is_time_density_matrices else FrequencyResultsCollector)
        calc_kwargs = dict(yield_total_ou=include_tcm, yield_proj_ou=include_tcm)

        out_fname = str(out_fname)
        if out_fname[-4:] == '.npz':
            writer = DipoleWriter(cls(self, calc_kwargs), only_one_pulse=only_one_pulse)
            writer.calculate_and_save_npz(out_fname, write_extra=write_extra)
        elif out_fname[-4:] == '.ulm':
            writer = DipoleWriter(cls(self, calc_kwargs, exclude=['dm_ouv']), only_one_pulse=only_one_pulse)
            writer.calculate_and_save_ulm(out_fname, write_extra=write_extra)
        else:
            raise ValueError(f'output-file must have ending .npz or .ulm, is {out_fname}')



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from typing import Any, Generator

from .base import BaseObservableCalculator
from ..density_matrices.base import WorkMetadata
from ..utils import ResultKeys, Result


class HotCarriersCalculator(BaseObservableCalculator):

    r""" Calculate hot-carrier distributions in the time domain.

    For weak perturbations, the response of the density matrix is
    to first order non-zero only in the occupied-unoccupied space,
    i.e. the block off-diagonals

    .. math::

        \delta\rho = \begin{bmatrix}
            0         &  [\delta\rho_{ai}^*] \\
            [\delta\rho_{ia}] &  0
        \end{bmatrix}.

    The unoccupied-occupied, or electron-hole, part of the density matrix is thus
    linear in perturbation and can by transformed using Fourier transforms.

    From the first-order response, the second order response, i.e. the hole-hole
    (:math:`\delta\rho_{ii'}`) and electron-electron (:math:`\delta\rho_{aa'}`) parts
    can be obtained.

    The hole-hole part is

    .. math::

        \delta\rho_{ii'} = - \frac{1}{2} \sum_n^{f_n > f_i, f_n > f_{i'}}
                           P_{ni} P_{ni'} + Q_{ni} Q_{ni'}

    and the electron-hole part

    .. math::

        \delta\rho_{aa'} = \frac{1}{2} \sum_n^{f_n < f_a, f_n < f_a'}
                           P_{ia} P_{ia'} + Q_{ia} Q_{ia'}

    where

    .. math::

        \begin{align}
            P_{ia} &= \frac{2 \mathrm{Im}\:\delta\rho_{ia}}{\sqrt{2 f_{ia}}} \\
            Q_{ia} &= \frac{2 \mathrm{Re}\:\delta\rho_{ia}}{\sqrt{2 f_{ia}}} ,
        \end{align}

    where :math:`f_{ia}` is the occupation number difference of pair :math:`ia`.

    Hot-carrier distributions are calculated by convolution of :math:`\delta\rho_{ii'}` (holes)
    and :math:`\delta\rho_{aa'}` (electrons) by Gaussian broadening functions on the energy
    grid.

    .. math::

        \begin{align}
            P^\text{holes}(\varepsilon) &=
            \sum_i \delta\rho_{ii'} G(\varepsilon - \varepsilon_i) \\
            P^\text{electrons}(\varepsilon) &=
            \sum_a \delta\rho_{aa'} G(\varepsilon - \varepsilon_a)
        \end{align}


    where :math:`\varepsilon_n` are Kohn-Sham energies and

    .. math::

       G(\varepsilon - \varepsilon_n)
        = \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            \left(\varepsilon_i - \varepsilon\right)^2
        }{
            2 \sigma^2
        }\right).

    Projected hot-carrier distributions are defined

    .. math::

        \begin{align}
            P^\text{holes}(\varepsilon) &=
            \sum_{ii'} \delta\rho_{ii'} w_{ii'} G(\varepsilon - \varepsilon_i) \\
            P^\text{electrons}(\varepsilon) &=
            \sum_{aa'} \delta\rho_{aa'} w_{aa'} G(\varepsilon - \varepsilon_a).
        \end{align}

    The Voronoi weights are

    .. math::

        W_{nn}
        = \left<\psi_n|\hat{w}|\psi_{n}\right>
        = \int w(\boldsymbol{r}) \psi_n^*(\boldsymbol{r}) \psi_{n}(\boldsymbol{r}) d\boldsymbol{r}

    where the operator :math:`\hat{w} = w(\boldsymbol{r})` is 1 in the Voronoi
    region of the atomic projections and 0 outside.

    Parameters
    ----------
    response
        Response object.
    voronoi
        Voronoi weights object.
    energies_occ
        Energy grid in units of eV for occupied levels (holes).
    energies_unocc
        Energy grid in units of eV for unoccupied levels (electrons)
    sigma
        Gaussian broadening width for energy grid in units of eV.
    times
        Compute hot carriers in the time domain, for these times (or as close to them as possible).
        In units of as.
    pulses
        Compute hot carriers in the time domain, in response to these pulses.
        If `None` no pulse convolution is performed.
    """

    def get_result_keys(self,
                        yield_total_hcdists: bool = False,
                        yield_proj_hcdists: bool = False,
                        yield_total_P: bool = False,
                        yield_proj_P: bool = False,
                        yield_total_P_ou: bool = False,
                        ):
        """ Get the keys that each result will contain, and dimensions thereof.

        Parameters
        ----------
        yield_total_hcdists
            The results should include the total hot-carrier distributions on the energy grid.
        yield_proj_hcdists
            The results should include the projections of the hot-carrier distributions on the energy grid.
        yield_total_P
            The results should include the total hot-carrier distributions in the electron-hole basis.
        yield_proj_P
            The results should include the projections of the hot-carrier distributions in the electron-hole basis.
        yield_total_P_ou
            The results should include the transition matrix broadened on the energy grid.
        """
        nI = self.nproj
        imin, imax, amin, amax = self.ksd.ialims()
        ni = int(imax - imin + 1)
        na = int(amax - amin + 1)
        no = len(self.energies_occ)
        nu = len(self.energies_unocc)

        resultkeys = ResultKeys('sumocc', 'sumunocc')

        if yield_total_P:
            resultkeys.add_key('P_i', ni)
            resultkeys.add_key('P_a', na)
        if yield_total_hcdists:
            resultkeys.add_key('hcdist_o', no)
            resultkeys.add_key('hcdist_u', nu)

        if yield_total_P_ou:
            resultkeys.add_key('P_ou', (no, nu))

        resultkeys.add_key('sumocc_proj_I', nI)
        resultkeys.add_key('sumunocc_proj_I', nI)
        if yield_proj_P:
            resultkeys.add_key('P_proj_Ii', (nI, ni))
            resultkeys.add_key('P_proj_Ia', (nI, na))
        if yield_proj_hcdists:
            resultkeys.add_key('hcdist_proj_Io', (nI, no))
            resultkeys.add_key('hcdist_proj_Iu', (nI, nu))

        return resultkeys

    def icalculate(self,
                   yield_total_hcdists: bool = False,
                   yield_proj_hcdists: bool = False,
                   yield_total_P: bool = False,
                   yield_proj_P: bool = False,
                   yield_total_P_ou: bool = False,
                   ) -> Generator[tuple[WorkMetadata, Result], None, None]:
        """ Iteratively calculate second order density matrices and hot-carrier distributions.

        Parameters
        ----------
        yield_total_hcdists
            The results should include the total hot-carrier distributions on the energy grid.
        yield_proj_hcdists
            The results should include the projections of the hot-carrier distributions on the energy grid.
        yield_total_P
            The results should include the total hot-carrier distributions in the electron-hole basis.
        yield_proj_P
            The results should include the projections of the hot-carrier distributions in the electron-hole basis.
        yield_total_P_ou
            The results should include the transition matrix broadened on the energy grid.

        Yields
        ------
        Tuple (work, result) on the root rank of the calculation communicator. \
        Does not yield on non-root ranks of the calculation communicator.

            work
                An object representing the metadata (time and pulse) for the work done.
            result
                Object containg the calculation results for this time and pulse.
        """
        include_energy_dists = (yield_proj_hcdists or yield_total_hcdists)
        if include_energy_dists:
            assert self.sigma is not None

        # List all keys that this method computes
        # This is necessary to safely send and receive data across ranks
        resultkeys = self.get_result_keys(yield_total_hcdists=yield_total_hcdists,
                                          yield_proj_hcdists=yield_proj_hcdists,
                                          yield_total_P=yield_total_P,
                                          yield_proj_P=yield_proj_P,
                                          yield_total_P_ou=yield_total_P_ou,
                                          )

        self._read_weights_eh()

        # Iterate over the pulses and times
        for work, dm in self.density_matrices:
            Q_ia = dm.Q_ia
            P_ia = dm.P_ia

            if dm.rank > 0:
                continue

            # Holes
            M_ii = 0.5 * (Q_ia @ Q_ia.T + P_ia @ P_ia.T)

            # Electrons
            M_aa = 0.5 * (Q_ia.T @ Q_ia + P_ia.T @ P_ia)

            result = Result()

            # (Optional) Compute broadened transition matrix
            if yield_total_P_ou:
                transition_ia = 0.5 * (Q_ia**2 + P_ia**2)
                result['P_ou'] = self.broaden_ia2ou(transition_ia)

            # Compute quantities in all space
            P_i = calculate_hcdist(None, M_ii)
            P_a = calculate_hcdist(None, M_aa)
            result['sumocc'] = np.sum(P_i)
            result['sumunocc'] = np.sum(P_a)
            if yield_total_hcdists:
                result['hcdist_o'] = self.broaden_occ(P_i)
                result['hcdist_u'] = self.broaden_unocc(P_a)
            if yield_total_P:
                result['P_i'] = P_i
                result['P_a'] = P_a

            result.create_all_empty(resultkeys)

            # Iterate over projections
            for iI, (weight_ii, weight_aa) in enumerate(self._iterate_weights_eh):
                P_proj_i = calculate_hcdist(weight_ii, M_ii)
                P_proj_a = calculate_hcdist(weight_aa, M_aa)
                result['sumocc_proj_I'][iI] = np.sum(P_proj_i)
                result['sumunocc_proj_I'][iI] = np.sum(P_proj_a)
                if yield_proj_hcdists:
                    result['hcdist_proj_Io'][iI] = self.broaden_occ(P_proj_i)
                    result['hcdist_proj_Iu'][iI] = self.broaden_unocc(P_proj_a)
                if yield_proj_P:
                    result['P_proj_Ii'][iI] = P_proj_i
                    result['P_proj_Ia'][iI] = P_proj_a

            self.log_parallel(f'Calculated and broadened HC distributions in {self.log.elapsed("calculate"):.2f}s '
                              f'for {work.desc}', flush=True)

            yield work, result

        if self.calc_comm.rank == 0 and self.density_matrices.localn > 0:
            self.log_parallel('Finished calculating hot-carrier matrices', flush=True)

    @property
    def _need_derivatives_real_imag(self) -> tuple[list[int], bool, bool]:
        return ([0], True, True)

    @property
    def _voronoi_shapes(self) -> dict[str, tuple[int, ...]]:
        nI = self.voronoi.nproj
        if nI == 0:
            return {}
        # Hole-hole part and electron-electron part
        Ni, Na = len(self.eig_i), len(self.eig_a)
        return {'ii': (nI, Ni, Ni), 'aa': (nI, Na, Na)}

    def calculate_totals_by_pulse_and_write(self,
                                            out_fname: str,
                                            write_extra: dict[str, Any] = dict()):
        """ Calculate the number of generated hot carriers, projected on groups of atoms, for
        a list of pulses.

        HC numbers are saved in a numpy archive if the file extension is `.npz`
        or in a comma separated file if the file extension is `.dat`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        """
        from ..writers.hcdist import HotCarriersWriter, write_hot_carrier_totals_by_pulse
        from ..writers.writer import PulseConvolutionAverageResultsCollector

        calc_kwargs = dict(yield_total_hcdists=False, yield_proj_hcdists=False)
        collector = PulseConvolutionAverageResultsCollector(self, calc_kwargs)
        writer = HotCarriersWriter(collector, only_one_pulse=False)

        out_fname = str(out_fname)
        if out_fname[-4:] == '.npz':
            writer.calculate_and_save_npz(out_fname, write_extra=write_extra)
        elif out_fname[-4:] == '.dat':
            write_hot_carrier_totals_by_pulse(out_fname, writer)
        else:
            raise ValueError(f'output-file must have ending .npz or .dat, is {out_fname}')

    def calculate_totals_by_time_and_write(self,
                                           out_fname: str,
                                           write_extra: dict[str, Any] = dict()):
        """ Calculate the number of generated hot carriers, projected on groups of atoms, for
        a list of times.

        HC numbers are saved in a numpy archive if the file extension is `.npz`
        or in a comma separated file if the file extension is `.dat`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        """
        from ..writers.hcdist import HotCarriersWriter, write_hot_carrier_totals_by_time
        from ..writers.writer import TimeResultsCollector

        calc_kwargs = dict(yield_total_hcdists=False, yield_proj_hcdists=False)
        collector = TimeResultsCollector(self, calc_kwargs)
        writer = HotCarriersWriter(collector, only_one_pulse=True)

        if out_fname[-4:] == '.npz':
            writer.calculate_and_save_npz(out_fname, write_extra=write_extra)
        elif out_fname[-4:] == '.dat':
            write_hot_carrier_totals_by_time(out_fname, writer)
        else:
            raise ValueError(f'output-file must have ending .npz or .dat, is {out_fname}')

    def calculate_distributions_and_write(self,
                                          out_fname: str,
                                          write_extra: dict[str, Any] = dict(),
                                          average_times: bool = True,
                                          only_one_pulse: bool = True):
        """ Calculate broadened hot-carrier energy distributions, optionally averaged over time.

        HC distributions are saved in a numpy archive if the file extension is `.npz`
        or in a comma separated file if the file extension is `.dat`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of extra key-value pairs to write to the data file.
        average_times
            If `True`, an average over the given times will be taken. If false, then
            hot-carrier distributions are computed separately over the times, and
            each output is written separately for each time.
        only_one_pulse
            There is only one pulse, don't group by pulse.
        """
        from ..writers.hcdist import HotCarriersWriter, write_hot_carrier_distributions
        from ..writers.writer import TimeResultsCollector, TimeAverageResultsCollector

        if len(self.energies_occ) == 0 and len(self.energies_unocc) == 0:
            raise ValueError('Either occupied or unoccupied energies grid must be given')

        calc_kwargs = dict(yield_total_hcdists=True, yield_proj_hcdists=True)
        cls = TimeAverageResultsCollector if average_times else TimeResultsCollector
        writer = HotCarriersWriter(cls(self, calc_kwargs), only_one_pulse=only_one_pulse)

        if out_fname[-4:] == '.npz':
            writer.calculate_and_save_npz(out_fname, write_extra=write_extra)
        elif out_fname[-4:] == '.dat':
            write_hot_carrier_distributions(out_fname, writer)
        else:
            raise ValueError(f'output-file must have ending .npz or .dat, is {out_fname}')


def calculate_hcdist(weight_xx: NDArray[np.float64] | None,
                     M_xx: NDArray[np.float64],
                     ) -> NDArray[np.float64]:
    r""" Calculate row-wise summed hot carrier distribution.

    .. math::

        \rho_n = \sum_{n'} \rho_{nn'} w_{nn'}

    Parameters
    ----------
    weight_xx
        Voronoi weights :math:`w_{ii}` or :math:`w_{aa}`.
        Specify None to let the weights be the identity matrix
    M_xx
        Matrix :math:`M_{ii}` or :math:`M_{aa}`

    Returns
    -------
        Hot carrier distribution by eigenvalue :math:`\rho_n`
    """
    if weight_xx is None:
        P_x = np.diag(M_xx)
    else:
        P_x = np.sum(weight_xx*M_xx, axis=0)

    return P_x
